"""Golden tests cases for testing liquid's built-in `split` filter."""

from liquid.golden.case import Case

cases = [
    Case(
        description="split string",
        template=r'{{ "Hi, how are you today?" | split: " " | join: "#" }}',
        expect="Hi,#how#are#you#today?",
    ),
    Case(
        description="not a string",
        template=r"{{ 56 | split: ' ' | first }}",
        expect="56",
    ),
    Case(
        description="argument not a string",
        template=r'{{ "hello th1ere" | split: 1 | join: "#" }}',
        expect="hello th#ere",
    ),
    Case(
        description="missing argument",
        template=r'{{ "hello there" | split }}',
        expect="",
        error=True,
    ),
    Case(
        description="too many arguments",
        template=r'{{ "hello there" | split: " ", "," }}',
        expect="",
        error=True,
    ),
    Case(
        description="undefined left value",
        template=r'{{ nosuchthing | split: " " }}',
        expect="",
    ),
    Case(
        description="undefined argument",
        template=r'{{ "Hello there" | split: nosuchthing | join: "#" }}',
        expect="H#e#l#l#o# #t#h#e#r#e",
    ),
]



"""Golden tests cases for testing liquid's built-in `newline_to_br` filter."""

from liquid.golden.case import Case

cases = [
    Case(
        description="string with newlines",
        template='{{ "- apples\n- oranges\n" | newline_to_br }}',
        expect="- apples<br />\n- oranges<br />\n",
    ),
    Case(
        description="not a string",
        template="{{ 5 | newline_to_br }}",
        expect="5",
    ),
    Case(
        description="unexpected argument",
        template='{{ "hello" | newline_to_br: 5 }}',
        expect="",
        error=True,
    ),
    Case(
        description="reference implementation test 1",
        template='{{ "a\nb\nc" | newline_to_br }}',
        expect="a<br />\nb<br />\nc",
    ),
    Case(
        description="reference implementation test 2",
        template='{{ "a\r\nb\nc" | newline_to_br }}',
        expect="a<br />\nb<br />\nc",
    ),
    Case(
        description="undefined left value",
        template="{{ nosuchthing | newline_to_br }}",
        expect="",
    ),
]



"""Shared configuration from which templates can be loaded and parsed."""

from __future__ import annotations
from functools import lru_cache
from pathlib import Path

from typing import Callable
from typing import ClassVar
from typing import Dict
from typing import Any
from typing import Type
from typing import Tuple
from typing import Union
from typing import Optional
from typing import Mapping
from typing import MutableMapping
from typing import TYPE_CHECKING

import warnings

from liquid.context import Undefined
from liquid.mode import Mode
from liquid.tag import Tag
from liquid.template import BoundTemplate
from liquid.lex import get_lexer
from liquid.stream import TokenStream
from liquid.parse import get_parser
from liquid.utils import LRUCache

from liquid.expressions import parse_boolean_expression
from liquid.expressions import parse_boolean_expression_with_parens
from liquid.expressions import parse_conditional_expression
from liquid.expressions import parse_conditional_expression_with_parens
from liquid.expressions import parse_filtered_expression
from liquid.expressions import parse_loop_expression

from liquid import ast
from liquid import builtin
from liquid import loaders

from liquid.exceptions import Error
from liquid.exceptions import LiquidSyntaxError
from liquid.exceptions import lookup_warning

if TYPE_CHECKING:  # pragma: no cover
    from liquid.expression import BooleanExpression
    from liquid.expression import FilteredExpression
    from liquid.expression import LoopExpression
    from liquid.context import Context


# pylint: disable=too-many-instance-attributes
class Environment:
    """Shared configuration from which templates can be loaded and parsed.

    An ``Environment`` is where you might register custom tags and filters, or store
    global context variables that should be included with every template.

    :param tag_start_string: The sequence of characters indicating the start of a
        liquid tag. Defaults to ``{%``.
    :type tag_start_string: str
    :param tag_end_string: The sequence of characters indicating the end of a liquid
        tag. Defaults to ``%}``.
    :type tag_end_string: str
    :param statement_start_string: The sequence of characters indicating the start of
        an output statement. Defaults to ``{{``.
    :type statement_start_string: str
    :param statement_end_string: The sequence of characters indicating the end of an
        output statement. Defaults to ``}}``
    :type statement_end_string: str
    :param template_comments: If ``True``, enable template comments. Where, by default,
        anything between ``{#`` and ``#}`` is considered a comment. Defaults to
        ``False``.
    :type template_comments: bool
    :param comment_start_string: The sequence of characters indicating the start of a
        comment. Defaults to ``{#``. ``template_comments`` must be ``True`` for
        ``comment_start_string`` to have any effect.
    :type comment_start_string: str
    :param comment_end_string: The sequence of characters indicating the end of a
        comment. Defaults to ``#}``. ``template_comments`` must be ``True`` for
        ``comment_end_string`` to have any effect.
    :type comment_end_string: str
    :param tolerance: Indicates how tolerant to be of errors. Must be one of
        ``Mode.LAX``, ``Mode.WARN`` or ``Mode.STRICT``. Defaults to ``Mode.STRICT``.
    :type tolerance: Mode
    :param loader: A template loader. If you want to use the builtin "render" or
        "include" tags, a loader must be configured. Defaults to an empty
        :class:`liquid.loaders.DictLoader`.
    :type loader: liquid.loaders.BaseLoader
    :param undefined: A subclass of :class:`Undefined` that represents undefined values.
        Could be one of the built-in undefined types, :class:`Undefined`,
        :class:`DebugUndefined` or :class:`StrictUndefined`. Defaults to
        :class:`Undefined`, an undefined type that silently ignores undefined values.
    :type undefined: liquid.Undefined
    :param strict_filters: If ``True``, will raise an exception upon finding an
        undefined filter. Otherwise undefined filters are silently ignored. Defaults to
        ``True``.
    :type strict_filters: bool
    :param autoescape: If `True`, all context values will be HTML-escaped before output
        unless they've been explicitly marked as "safe". Requires the package
        Markupsafe. Defaults to ``False``.
    :type autoescape: bool
    :param auto_reload: If `True`, loaders that have an ``uptodate`` callable will
        reload template source data automatically. For deployments where template
        sources don't change between service reloads, setting auto_reload to `False` can
        yield an increase in performance by avoiding calls to ``uptodate``. Defaults to
        ``True``.
    :type auto_reload: bool
    :param cache_size: The capacity of the template cache in number of templates.
        Defaults to 300. If ``cache_size`` is ``None`` or less than ``1``, it has the
        effect of setting ``auto_reload`` to ``False``.
    :type cache_size: int
    :param expression_cache_size: The capacity of each of the common expression caches.
        Defaults to ``0``, disabling expression caching.
    :type expression_cache_size: int
    :param globals: An optional mapping that will be added to the context of any
        template loaded from this environment. Defaults to ``None``.
    :type globals: dict
    """

    # Maximum number of times a context can be extended or wrapped before raising
    # a ContextDepthError.
    context_depth_limit: ClassVar[int] = 30

    # Maximum number of loop iterations allowed before a LoopIterationLimitError is
    # raised.
    loop_iteration_limit: ClassVar[Optional[int]] = None

    # Maximum number of bytes (according to sys.getsizeof) allowed in a template's
    # local namespace before a LocalNamespaceLimitError is raised. We only count the
    # size of the namespaces values, not the size of keys/names.
    local_namespace_limit: ClassVar[Optional[int]] = None

    # Maximum number of bytes that can be written to a template's output stream before
    # raising an OutputStreamLimitError.
    output_stream_limit: ClassVar[Optional[int]] = None

    # Instances of ``template_class`` are returned from ``from_string``,
    # ``get_template`` and ``get_template_async``. It should be the ``BoundTemplate``
    # class or a subclass of it.
    template_class = BoundTemplate

    # pylint: disable=redefined-builtin too-many-arguments too-many-locals
    def __init__(
        self,
        tag_start_string: str = r"{%",
        tag_end_string: str = r"%}",
        statement_start_string: str = r"{{",
        statement_end_string: str = r"}}",
        strip_tags: bool = False,
        tolerance: Mode = Mode.STRICT,
        loader: Optional[loaders.BaseLoader] = None,
        undefined: Type[Undefined] = Undefined,
        strict_filters: bool = True,
        autoescape: bool = False,
        auto_reload: bool = True,
        cache_size: int = 300,
        globals: Optional[Mapping[str, object]] = None,
        template_comments: bool = False,
        comment_start_string: str = "{#",
        comment_end_string: str = "#}",
        expression_cache_size: int = 0,
    ):
        self.tag_start_string = tag_start_string
        self.tag_end_string = tag_end_string
        self.statement_start_string = statement_start_string
        self.statement_end_string = statement_end_string

        # Automatic tag stripping is not yet implemented. Changing this has no effect.
        self.strip_tags = strip_tags

        # An instance of a template loader implementing ``liquid.loaders.BaseLoader``.
        # ``get_template()`` will delegate to this loader.
        self.loader = loader or loaders.DictLoader({})

        # A mapping of template variable names to python objects. These variables will
        # be added to the global namespace of any template rendered from this
        # environment using ``from_string`` or ``get_template``.
        self.globals: Mapping[str, object] = globals or {}

        # Extended template comment syntax control.
        self.template_comments = template_comments
        self.comment_start_string = comment_start_string
        self.comment_end_string = comment_end_string
        if not self.template_comments:
            self.comment_start_string = ""
            self.comment_end_string = ""

        # The undefined type. When an identifier can not be resolved, the returned value
        # is ``Undefined`` or a subclass of ``Undefined``.
        self.undefined = undefined

        # Indicates if an undefined filter should raise an exception or be ignored.
        self.strict_filters = strict_filters

        # Indicates if autoescape is enabled.
        self.autoescape = autoescape

        # Tag register.
        self.tags: Dict[str, Tag] = {}

        # Filter register.
        self.filters: Dict[str, Callable[..., Any]] = {}

        # tolerance mode
        self.mode = tolerance

        # Template cache
        if cache_size and cache_size > 0:
            self.cache: MutableMapping[Any, Any] = LRUCache(cache_size)
            self.auto_reload = auto_reload
        else:
            self.cache = {}
            self.auto_reload = False

        # Common expression parsing functions that might be cached.
        self.expression_cache_size = expression_cache_size
        (
            self.parse_boolean_expression_value,
            self.parse_boolean_expression_value_with_parens,
            self.parse_conditional_expression_value,
            self.parse_conditional_expression_value_with_parens,
            self.parse_filtered_expression_value,
            self.parse_loop_expression_value,
        ) = self._get_expression_parsers(self.expression_cache_size)

        builtin.register(self)

    def __hash__(self) -> int:
        return hash(
            (
                self.statement_start_string,
                self.statement_end_string,
                self.tag_start_string,
                self.tag_end_string,
                self.comment_start_string,
                self.comment_end_string,
                self.mode,
                self.strip_tags,
                # Necessary when replacing the standard output statement implementation.
                self.tags.get("statement"),
            )
        )

    def add_tag(self, tag: Type[Tag]) -> None:
        """Register a liquid tag with the environment. Built-in tags are registered for
        you automatically with every new :class:`Environment`.

        :param tag: The tag to register.
        :type tag: Type[Tag]
        """
        self.tags[tag.name] = tag(self)

    def add_filter(self, name: str, func: Callable[..., Any]) -> None:
        """Register a filter function with the environment.

        :param name: The filter's name. Does not need to match the function name. This
            is what you'll use to apply the filter to an expression in a liquid
            template.
        :type name: str
        :param func: Any callable that accepts at least one argument, the result of the
            expression the filter is applied to. If the filter needs access to the
            environment or render context, you probably want to make ``func`` a class
            that inherits from :class:`liquid.filter.Filter`, and override the
            ``__call__`` method. All builtin filters are implemented in this way.
        :type func: Callable[..., Any]
        """
        self.filters[name] = func

    def parse(self, source: str) -> ast.ParseTree:
        """Parse the given string as a liquid template.

        More often than not you'll want to use `Environment.from_string` instead.
        """
        tokenize = get_lexer(
            self.tag_start_string,
            self.tag_end_string,
            self.statement_start_string,
            self.statement_end_string,
            self.comment_start_string,
            self.comment_end_string,
        )

        parser = get_parser(self)
        token_iter = tokenize(source)
        return parser.parse(TokenStream(token_iter))

    # pylint: disable=redefined-builtin
    def from_string(
        self,
        source: str,
        name: str = "",
        path: Optional[Union[str, Path]] = None,
        globals: Optional[Mapping[str, object]] = None,
        matter: Optional[Mapping[str, object]] = None,
    ) -> BoundTemplate:
        """Parse the given string as a liquid template.

        :param source: The liquid template source code.
        :type source: str
        :param name: Optional name of the template. Available as ``Template.name``.
            Defaults to the empty string.
        :type name: str
        :param path: Optional path or identifier to the origin of the template. Defaults
            to ``None``.
        :type path: str, pathlib.Path
        :param globals: An optional mapping of context variables made available every
            time the resulting template is rendered. Defaults to ``None``.
        :type globals: dict
        :param matter: Optional mapping containing variables associated with the
            template. Could be "front matter" or other meta data.
        :type matter: Optional[Mapping[str, object]]
        :returns: A parsed template ready to be rendered.
        :rtype: liquid.template.BoundTemplate
        """
        try:
            parse_tree = self.parse(source)
        except LiquidSyntaxError as err:
            err.filename = path
            err.source = source
            raise err
        except Exception as err:
            raise Error("unexpected liquid parsing error") from err
        return self.template_class(
            env=self,
            name=name,
            path=path,
            parse_tree=parse_tree,
            globals=self.make_globals(globals),
            matter=matter,
        )

    # pylint: disable=redefined-builtin
    def get_template(
        self,
        name: str,
        globals: Optional[Mapping[str, object]] = None,
    ) -> BoundTemplate:
        """Load and parse a template using the configured loader.

        :param name: The template's name. The loader is responsible for interpretting
            the name.
        :param globals: A mapping of context variables made available every time the
            resulting template is rendered.
        :returns: A parsed template ready to be rendered.
        :rtype: liquid.template.BoundTemplate
        :raises:
            :class:`liquid.exceptions.TemplateNotFound`: if a template with the given
            name can not be found.
        """
        template = self._check_cache(name, globals)

        if not template:
            template = self.loader.load(self, name, globals=self.make_globals(globals))
            self.cache[name] = template

        return template

    async def get_template_async(
        self,
        name: str,
        globals: Optional[Mapping[str, object]] = None,
    ) -> BoundTemplate:
        """An async version of ``get_template``."""
        template = self.cache.get(name)

        if isinstance(template, BoundTemplate) and (
            not self.auto_reload or await template.is_up_to_date_async()
        ):
            template.globals.update(self.make_globals(globals))
        else:
            template = await self.loader.load_async(
                self,
                name,
                globals=self.make_globals(globals),
            )
            self.cache[name] = template

        return template

    def get_template_with_context(
        self,
        context: "Context",
        name: str,
        **kwargs: str,
    ) -> BoundTemplate:
        """Load and parse a template using the configured loader, optionally referencing
        a render context."""
        # No template caching. How would we know what context variables a loader needs?
        # A custom loader that uses context could implement its own cache.
        return self.loader.load_with_context(context, name, **kwargs)

    async def get_template_with_context_async(
        self,
        context: "Context",
        name: str,
        **kwargs: str,
    ) -> BoundTemplate:
        """An async version of ``get_template_with_context``."""
        return await self.loader.load_with_context_async(context, name, **kwargs)

    def _check_cache(
        self,
        name: str,
        globals: Optional[Mapping[str, object]] = None,
    ) -> Optional[BoundTemplate]:
        _cached = self.cache.get(name)

        if isinstance(_cached, BoundTemplate) and (
            not self.auto_reload or _cached.is_up_to_date
        ):
            _cached.globals.update(self.make_globals(globals))
            return _cached
        return None

    # pylint: disable=redefined-builtin
    def make_globals(
        self, globals: Optional[Mapping[str, object]] = None
    ) -> Dict[str, object]:
        """Combine environment globals with template globals."""
        if globals:
            # Template globals take priority over environment globals.
            return {**self.globals, **globals}
        return dict(self.globals)

    def error(
        self,
        exc: Union[Type[Error], Error],
        msg: Optional[str] = None,
        linenum: Optional[int] = None,
    ) -> None:
        """Raise, warn or ignore the given exception according to the current mode."""
        if not isinstance(exc, Error):
            exc = exc(msg, linenum=linenum)
        elif not exc.linenum:
            exc.linenum = linenum

        if self.mode == Mode.STRICT:
            raise exc
        if self.mode == Mode.WARN:
            warnings.warn(str(exc), category=lookup_warning(exc.__class__))

    def set_expression_cache_size(self, maxsize: int = 0) -> None:
        """Create or replace cached versions of the common expression parsers. If
        `maxsize` is less than ``1``, no expression caching will happen.

        :param maxsize: The maximum size of each expression cache.
        :type maxsize: int
        """
        self.expression_cache_size = maxsize
        (
            self.parse_boolean_expression_value,
            self.parse_boolean_expression_value_with_parens,
            self.parse_conditional_expression_value,
            self.parse_conditional_expression_value_with_parens,
            self.parse_filtered_expression_value,
            self.parse_loop_expression_value,
        ) = self._get_expression_parsers(self.expression_cache_size)

    def _get_expression_parsers(
        self, cache_size: int = 0
    ) -> Tuple[
        Callable[[str], "BooleanExpression"],
        Callable[[str], "BooleanExpression"],
        Callable[[str], "FilteredExpression"],
        Callable[[str], "FilteredExpression"],
        Callable[[str], "FilteredExpression"],
        Callable[[str], "LoopExpression"],
    ]:
        if cache_size >= 1:
            return (
                lru_cache(maxsize=cache_size)(parse_boolean_expression),
                lru_cache(maxsize=cache_size)(parse_boolean_expression_with_parens),
                lru_cache(maxsize=cache_size)(parse_conditional_expression),
                lru_cache(maxsize=cache_size)(parse_conditional_expression_with_parens),
                lru_cache(maxsize=cache_size)(parse_filtered_expression),
                lru_cache(maxsize=cache_size)(parse_loop_expression),
            )
        return (
            parse_boolean_expression,
            parse_boolean_expression_with_parens,
            parse_conditional_expression,
            parse_conditional_expression_with_parens,
            parse_filtered_expression,
            parse_loop_expression,
        )


# pylint: disable=redefined-builtin too-many-arguments too-many-locals
@lru_cache(maxsize=10)
def get_implicit_environment(
    tag_start_string: str,
    tag_end_string: str,
    statement_start_string: str,
    statement_end_string: str,
    strip_tags: bool,
    tolerance: Mode,
    loader: Optional[loaders.BaseLoader],
    undefined: Type[Undefined],
    strict_filters: bool,
    autoescape: bool,
    auto_reload: bool,
    cache_size: int,
    globals: Optional[Mapping[str, object]],
    template_comments: bool,
    comment_start_string: str,
    comment_end_string: str,
    expression_cache_size: int,
) -> Environment:
    """Return an :class:`Environment` initialized with the given arguments."""
    return Environment(
        tag_start_string=tag_start_string,
        tag_end_string=tag_end_string,
        statement_start_string=statement_start_string,
        statement_end_string=statement_end_string,
        strip_tags=strip_tags,
        tolerance=tolerance,
        loader=loader,
        undefined=undefined,
        strict_filters=strict_filters,
        autoescape=autoescape,
        auto_reload=auto_reload,
        cache_size=cache_size,
        globals=globals,
        template_comments=template_comments,
        comment_start_string=comment_start_string,
        comment_end_string=comment_end_string,
        expression_cache_size=expression_cache_size,
    )


# ``Template`` is a factory function masquerading as a class. The desire to have an
# intuitive API and to please the static type checker outweighs this abuse of Python
# naming conventions. At least for now.

# pylint: disable=redefined-builtin too-many-arguments invalid-name too-many-locals
def Template(
    source: str,
    tag_start_string: str = r"{%",
    tag_end_string: str = r"%}",
    statement_start_string: str = r"{{",
    statement_end_string: str = r"}}",
    strip_tags: bool = False,
    tolerance: Mode = Mode.STRICT,
    undefined: Type[Undefined] = Undefined,
    strict_filters: bool = True,
    autoescape: bool = False,
    auto_reload: bool = True,
    cache_size: int = 300,
    globals: Optional[Mapping[str, object]] = None,
    template_comments: bool = False,
    comment_start_string: str = "{#",
    comment_end_string: str = "#}",
    expression_cache_size: int = 0,
) -> BoundTemplate:
    """Returns a :class:`liquid.template.BoundTemplate`, automatically creating an
    :class:`Environment` to bind it to.

    Other than `source`, all arguments are passed to the implicit :class:`Environment`,
    which might have been cached from previous calls to :func:`Template`.

    :param source: A Liquid template source string.
    :type source: str
    :param tag_start_string: The sequence of characters indicating the start of a
        liquid tag. Defaults to ``{%``.
    :type tag_start_string: str
    :param tag_end_string: The sequence of characters indicating the end of a liquid
        tag. Defaults to ``%}``.
    :type tag_end_string: str
    :param statement_start_string: The sequence of characters indicating the start of
        an output statement. Defaults to ``{{``.
    :type statement_start_string: str
    :param statement_end_string: The sequence of characters indicating the end of an
        output statement. Defaults to ``}}``
    :type statement_end_string: str
    :param template_comments: If ``True``, enable template comments. Where, by default,
        anything between ``{#`` and ``#}`` is considered a comment. Defaults to
        ``False``.
    :type template_comments: bool
    :param comment_start_string: The sequence of characters indicating the start of a
        comment. Defaults to ``{#``. ``template_comments`` must be ``True`` for
        ``comment_start_string`` to have any effect.
    :type comment_start_string: str
    :param comment_end_string: The sequence of characters indicating the end of a
        comment. Defaults to ``#}``. ``template_comments`` must be ``True`` for
        ``comment_end_string`` to have any effect.
    :type comment_end_string: str
    :param tolerance: Indicates how tolerant to be of errors. Must be one of
        `Mode.LAX`, `Mode.WARN` or `Mode.STRICT`. Defaults to ``Mode.STRICT``.
    :type tolerance: Mode
    :param undefined: A subclass of :class:`Undefined` that represents undefined values.
        Could be one of the built-in undefined types, :class:`Undefined`,
        :class:`DebugUndefined` or :class:`StrictUndefined`. Defaults to
        :class:`Undefined`, an undefined type that silently ignores undefined values.
    :type undefined: Undefined
    :param strict_filters: If ``True``, will raise an exception upon finding an
        undefined filter. Otherwise undefined filters are silently ignored. Defaults to
        ``True``.
    :type strict_filters: bool
    :param autoescape: If `True`, all context values will be HTML-escaped before output
        unless they've been explicitly marked as "safe". Requires the package
        Markupsafe. Defaults to ``False``.
    :type autoescape: bool
    :param auto_reload: If `True`, loaders that have an ``uptodate`` callable will
        reload template source data automatically. For deployments where template
        sources don't change between service reloads, setting auto_reload to `False` can
        yield an increase in performance by avoiding calls to ``uptodate``. Defaults to
        ``True``.
    :type auto_reload: bool
    :param cache_size: The capacity of the template cache in number of templates.
        Defaults to 300. If ``cache_size`` is ``None`` or less than ``1``, it has the
        effect of setting ``auto_reload`` to ``False``.
    :type cache_size: int
    :param expression_cache_size: The capacity of each of the common expression caches.
        Defaults to ``0``, disabling expression caching.
    :type expression_cache_size: int
    :param globals: An optional mapping that will be added to the context of any
        template loaded from this environment. Defaults to ``None``.
    :type globals: dict
    """
    # Resorting to named arguments (repeated 3 times) as I've twice missed a bug
    # because of positional arguments.
    env = get_implicit_environment(
        tag_start_string=tag_start_string,
        tag_end_string=tag_end_string,
        statement_start_string=statement_start_string,
        statement_end_string=statement_end_string,
        strip_tags=strip_tags,
        tolerance=tolerance,
        loader=None,
        undefined=undefined,
        strict_filters=strict_filters,
        autoescape=autoescape,
        auto_reload=auto_reload,
        cache_size=cache_size,
        globals=None,
        template_comments=template_comments,
        comment_start_string=comment_start_string,
        comment_end_string=comment_end_string,
        expression_cache_size=expression_cache_size,
    )

    return env.from_string(source, globals=globals)



"""Golden tests cases for testing liquid output statements."""

import datetime
from liquid.golden.case import Case

cases = [
    Case(
        description="render a string literal",
        template=r"{{ 'hello' }}",
        expect="hello",
    ),
    Case(
        description="render an integer literal",
        template=r"{{ 123 }}",
        expect="123",
    ),
    Case(
        description="render a negative integer literal",
        template=r"{{ -123 }}",
        expect="-123",
    ),
    Case(
        description="render a float literal",
        template=r"{{ 1.23 }}",
        expect="1.23",
    ),
    Case(
        description="render nil",
        template=r"{{ nil }}",
        expect="",
    ),
    Case(
        description="render a variable from the global namespace",
        template=r"{{ product.title }}",
        expect="foo",
        globals={"product": {"title": "foo"}},
    ),
    Case(
        description="render a variable from the local namespace",
        template=r"{% assign name = 'Brian' %}{{ name }}",
        expect="Brian",
    ),
    Case(
        description="render an undefined variable",
        template=r"{{ age }}",
        expect="",
    ),
    Case(
        description="render an undefined property",
        template=r"{{ product.age }}",
        expect="",
        globals={"product": {"title": "foo"}},
    ),
    Case(
        description="access an array item by index",
        template=r"{{ product.tags[1] }}",
        expect="garden",
        globals={"product": {"tags": ["sports", "garden"]}},
    ),
    Case(
        description="access an array item by negative index",
        template=r"{{ product.tags[-2] }}",
        expect="sports",
        globals={"product": {"tags": ["sports", "garden"]}},
    ),
    Case(
        description="access array item by index stored in a local variable",
        template="{% assign i = 1 %}{{ product.tags[i] }}",
        expect="garden",
        globals={"product": {"tags": ["sports", "garden"]}},
    ),
    Case(
        description="render a global variable with a filter",
        template=r"{{ product.title | upcase }}",
        expect="FOO",
        globals={"product": {"title": "foo"}},
    ),
    Case(
        description="dump an array from the global context",
        template=r"{{ product.tags | join: '#' }}",
        expect="sports#garden",
        globals={"product": {"tags": ["sports", "garden"]}},
    ),
    Case(
        description="assign a variable the value of an existing variable",
        template=(
            r"{% capture some %}hello{% endcapture %}"
            r"{% assign other = some %}"
            r"{% assign some = 'foo' %}"
            r"{{ some }}-{{ other }}"
        ),
        expect="foo-hello",
    ),
    Case(
        description="traverse variables with bracketed identifiers",
        template=r"{{ site.data.menu[include.menu][include.locale] }}",
        expect="it works!",
        globals={
            "site": {"data": {"menu": {"foo": {"bar": "it works!"}}}},
            "include": {"menu": "foo", "locale": "bar"},
        },
    ),
    Case(
        description="render the special built in variable 'now'",
        template=r"{{ now | date: '%d/%m/%Y' }}",
        expect=datetime.datetime.now().strftime(r"%d/%m/%Y"),
        standard=False,
    ),
    Case(
        description="render the special built in variable 'today'",
        template=r"{{ today | date: '%d/%m/%Y' }}",
        expect=datetime.date.today().strftime(r"%d/%m/%Y"),
        standard=False,
    ),
    Case(
        description="render a default given a literal false",
        template=r"{{ false | default: 'bar' }}",
        expect="bar",
    ),
    Case(
        description=(
            "render a default given a literal false " "with 'allow false' equal to true"
        ),
        template=r"{{ false | default: 'bar', allow_false: true }}",
        expect="false",
    ),
    Case(
        description=(
            "render a default given a literal false "
            "with 'allow false' equal to false"
        ),
        template=r"{{ false | default: 'bar', allow_false: false }}",
        expect="bar",
    ),
    Case(
        description=("unexpected left value for the `join` filter passes through"),
        template=r"{{ 12 | join: '#' }}",
        expect="12",
    ),
    Case(
        description="render an output start sequence as a string literal",
        template=r"{{ '{{' }}",
        expect=r"{{",
    ),
    Case(
        description="access an undefined variable by index",
        template=r"{{ nosuchthing[0] }}",
        expect="",
    ),
    Case(
        description="render a range object",
        template=r"{{ (1..5) | join: '#' }}",
        expect="1#2#3#4#5",
    ),
    Case(
        description="render a range object that uses a float",
        template=r"{{ (1.4..5) | join: '#' }}",
        expect="1#2#3#4#5",
    ),
    Case(
        description="render a range object that uses an identifier",
        template=r"{{ (foo..5) | join: '#' }}",
        expect="2#3#4#5",
        globals={"foo": 2},
    ),
    Case(
        description="reverse a range",
        template=r"{{ (foo..5) | reverse | join: '#' }}",
        expect="5#4#3#2",
        globals={"foo": 2},
    ),
    Case(
        description="chained bracketed identifier index",
        template=r"{{ products[0].title }}",
        expect="shoe",
        globals={"products": [{"title": "shoe"}, {"title": "hat"}]},
    ),
    Case(
        description="chained bracketed identifier index no dot",
        template=r"{{ products[0]title }}",
        expect="shoe",
        globals={"products": [{"title": "shoe"}, {"title": "hat"}]},
    ),
    Case(
        description="chained identifier dot separated index",
        template=r"{{ products.0.title }}",
        expect="",
        globals={"products": [{"title": "shoe"}, {"title": "hat"}]},
        strict=True,
        error=True,
    ),
]



"""Filter function helpers."""

from __future__ import annotations

from functools import wraps

from typing import Any
from typing import Callable
from typing import Iterable
from typing import Iterator
from typing import List
from typing import Union
from typing import Optional
from typing import TYPE_CHECKING

from liquid.context import Undefined

from liquid.exceptions import FilterArgumentError
from liquid.exceptions import FilterValueError

from liquid.limits import to_int

if TYPE_CHECKING:  # pragma: no cover
    FilterT = Callable[..., Any]
    NumberT = Union[float, int]


def with_context(_filter: FilterT) -> FilterT:
    """Pass the active :class:`liquid.context.Context` as the named argument
    ``context`` to the decorated filter function.

    :param _filter: The filter function to decorate.
    :type _filter: Callable[..., Any]
    """
    _filter.with_context = True  # type: ignore
    return _filter


def with_environment(_filter: FilterT) -> FilterT:
    """Pass the active :class:`liquid.Environment` as the named argument
    ``environment`` to the decorated filter function.

    :param _filter: The filter function to decorate.
    :type _filter: Callable[..., Any]
    """
    _filter.with_environment = True  # type: ignore
    return _filter


def string_filter(_filter: FilterT) -> FilterT:
    """A filter function decorator that converts the first positional argument to a
    string."""

    @wraps(_filter)
    def wrapper(val: object, *args: Any, **kwargs: Any) -> Any:
        if not isinstance(val, str):
            val = str(val)
        try:
            return _filter(val, *args, **kwargs)
        except TypeError as err:
            raise FilterArgumentError(err) from err

    return wrapper


def array_filter(_filter: FilterT) -> FilterT:
    """A filter function decorator that raises a FilterValueError if the filter value
    is not array-like."""

    @wraps(_filter)
    def wrapper(val: object, *args: Any, **kwargs: Any) -> Any:
        if not isinstance(val, (list, tuple, Undefined, range)):
            raise FilterValueError(f"expected an array, found {type(val).__name__}")

        try:
            return _filter(val, *args, **kwargs)
        except TypeError as err:
            raise FilterArgumentError(err) from err

    return wrapper


def sequence_filter(_filter: FilterT) -> FilterT:
    """A filter function decorator that raises a FilterValueError if the filter value
    can not be coerced into an array-like object.

    This is intended to mimic the semantics of the reference implementation's
    ``InputIterator`` class.
    """

    @wraps(_filter)
    def wrapper(val: object, *args: Any, **kwargs: Any) -> Any:
        if isinstance(val, (list, tuple)):
            val = flatten(val)
        elif isinstance(val, (dict, str)) or not isinstance(val, Iterable):
            val = [val]

        try:
            return _filter(val, *args, **kwargs)
        except TypeError as err:
            raise FilterArgumentError(err) from err

    return wrapper


def liquid_filter(_filter: FilterT) -> FilterT:
    """A filter function decorator that wraps `TypeError` in `FilterArgumentError`."""

    @wraps(_filter)
    def wrapper(val: object, *args: Any, **kwargs: Any) -> Any:
        try:
            return _filter(val, *args, **kwargs)
        except TypeError as err:
            raise FilterArgumentError(err) from err

    return wrapper


def int_arg(val: Any, default: Optional[int] = None) -> int:
    """Return the ``val`` as an int or ``default`` if ``val`` can't be cast to an
    int."""
    try:
        return to_int(val)
    except ValueError as err:
        if default is not None:
            return default
        raise FilterArgumentError(
            f"expected an int or string, found {type(val).__name__}"
        ) from err


def num_arg(val: Any, default: Optional[NumberT] = None) -> NumberT:
    """Return the ``val`` as an int or float. If ``val`` can't be cast to an
    int or float, return ``default`."""
    if isinstance(val, (int, float)):
        return val

    if isinstance(val, str):
        try:
            return to_int(val)
        except ValueError:
            pass

        try:
            return float(val)
        except ValueError as err:
            if default is not None:
                return default
            raise FilterArgumentError(
                f"could not cast string '{val}' to a number"
            ) from err

    elif default is not None:
        return default

    raise FilterArgumentError(
        f"expected an int, float or string, found {type(val).__name__}"
    )


def math_filter(_filter: FilterT) -> FilterT:
    """A filter function decorator that raises a FilterArgumentError if the filter value
    is not a number."""

    @wraps(_filter)
    def wrapper(val: object, *args: Any, **kwargs: Any) -> Any:
        val = num_arg(val, default=0)

        try:
            return _filter(val, *args, **kwargs)
        except TypeError as err:
            raise FilterArgumentError(err) from err

    return wrapper


def flatten(it: Iterable[Any], level: int = 5) -> List[object]:
    """Flatten nested "liquid arrays" into a list."""

    def _flatten(it: Iterable[Any], level: int = 5) -> Iterator[object]:
        for obj in it:
            if not level or not isinstance(obj, (list, tuple)):
                yield obj
            else:
                yield from _flatten(obj, level=level - 1)

    return list(_flatten(it, level))



"""Golden tests cases for testing liquid's built-in `date` filter."""

from liquid.golden.case import Case

cases = [
    Case(
        description="well formed string",
        template=r"{{ 'March 14, 2016' | date: '%b %d, %y' }}",
        expect="Mar 14, 16",
    ),
    Case(
        description="too many arguments",
        template=r"{{ 'March 14, 2016' | date: '%b %d, %y', 'foo' }}",
        expect="",
        error=True,
    ),
    Case(
        description="undefined left value",
        template=r"{{ nosuchthing | date: '%b %d, %y' }}",
        expect="",
    ),
    Case(
        description="missing argument",
        template=r"{{ 'March 14, 2016' | date }}",
        expect="",
        error=True,
    ),
    Case(
        description="undefined argument",
        template=r"{{ 'March 14, 2016' | date: nosuchthing }}",
        expect="March 14, 2016",
    ),
    Case(
        description="literal percent",
        template=r"{{ 'March 14, 2016' | date: '%%%b %d, %y' }}",
        expect="%Mar 14, 16",
    ),
    Case(
        description="timestamp integer",
        template=r"{{ 1152098955 | date: '%m/%d/%Y' }}",
        expect="07/05/2006",
    ),
    Case(
        description="timestamp string",
        template=r"{{ '1152098955' | date: '%m/%d/%Y' }}",
        expect="07/05/2006",
    ),
    # Case(
    #     description="negative timestamp integer",
    #     template=r"{{ -1152098955 | date: '%m/%d/%Y' }}",
    #     expect="06/29/1933",
    # ),
    Case(
        description="negative timestamp string",
        template=r"{{ '-1152098955' | date: '%m/%d/%Y' }}",
        expect="-1152098955",
    ),
]



"""Test miscellaneous filter functions."""
# pylint: disable=too-many-public-methods,too-many-lines,missing-class-docstring
import datetime
import decimal
import platform
import unittest

from functools import partial
from inspect import isclass

from typing import NamedTuple
from typing import Any
from typing import List
from typing import Dict

from liquid.environment import Environment

from liquid.exceptions import Error
from liquid.exceptions import FilterArgumentError

from liquid.expression import EMPTY

from liquid.builtin.filters.misc import size
from liquid.builtin.filters.misc import default
from liquid.builtin.filters.misc import date


class Case(NamedTuple):
    description: str
    val: Any
    args: List[Any]
    kwargs: Dict[Any, Any]
    expect: Any


class MockDrop:  # pragma: no cover
    def __init__(self, val):
        self.val = val

    def __eq__(self, other):
        if isinstance(other, MockDrop) and self.val == other.val:
            return True
        return False

    def __str__(self):
        return "hello mock drop"

    def __liquid__(self):
        return self.val


class NoLiquidDrop:  # pragma: no cover
    def __init__(self, val):
        self.val = val

    def __eq__(self, other):
        if isinstance(other, NoLiquidDrop) and self.val == other.val:
            return True
        return False

    def __str__(self):
        return "hello no liquid drop"


class FalsyDrop:  # pragma: no cover
    def __init__(self, val):
        self.val = val

    def __eq__(self, other):
        if isinstance(other, bool) and self.val == other:
            return True
        if isinstance(other, FalsyDrop) and self.val == other.val:
            return True
        return False

    def __str__(self):
        return "falsy drop"

    def __bool__(self):
        return False


class MiscFilterTestCase(unittest.TestCase):
    """Test miscellaneous filter functions."""

    def setUp(self) -> None:
        self.env = Environment()

    def _test(self, func, test_cases):
        if getattr(func, "with_environment", False):
            func = partial(func, environment=self.env)

        for case in test_cases:
            with self.subTest(msg=case.description):
                if isclass(case.expect) and issubclass(case.expect, Error):
                    with self.assertRaises(case.expect):
                        func(case.val, *case.args, **case.kwargs)
                else:
                    self.assertEqual(
                        func(case.val, *case.args, **case.kwargs), case.expect
                    )

    def test_size(self):
        """Test `size` filter function."""
        test_cases = [
            Case(
                description="array",
                val=["a", "b", "c"],
                args=[],
                kwargs={},
                expect=3,
            ),
            Case(
                description="string",
                val="abc",
                args=[],
                kwargs={},
                expect=3,
            ),
            Case(
                description="empty array",
                val=[],
                args=[],
                kwargs={},
                expect=0,
            ),
            Case(
                description="unexpected argument",
                val=[1, 2, 3],
                args=["foo"],
                kwargs={},
                expect=FilterArgumentError,
            ),
            Case(
                description="not an array or string",
                val=1,
                args=[],
                kwargs={},
                expect=0,
            ),
            Case(
                description="undefined left value",
                val=self.env.undefined("test"),
                args=[],
                kwargs={},
                expect=0,
            ),
        ]

        self._test(size, test_cases)

    def test_default(self):
        """Test `default` filter function."""
        test_cases = [
            Case(
                description="nil",
                val=None,
                args=["foo"],
                kwargs={},
                expect="foo",
            ),
            Case(
                description="false",
                val=False,
                args=["foo"],
                kwargs={},
                expect="foo",
            ),
            Case(
                description="empty string",
                val="",
                args=["foo"],
                kwargs={},
                expect="foo",
            ),
            Case(
                description="empty list",
                val=[],
                args=["foo"],
                kwargs={},
                expect="foo",
            ),
            Case(
                description="empty object",
                val={},
                args=["foo"],
                kwargs={},
                expect="foo",
            ),
            Case(
                description="not empty string",
                val="hello",
                args=["foo"],
                kwargs={},
                expect="hello",
            ),
            Case(
                description="not empty list",
                val=["hello"],
                args=["foo"],
                kwargs={},
                expect=["hello"],
            ),
            Case(
                description="not empty object",
                val={"title": "hello"},
                args=["foo"],
                kwargs={},
                expect={"title": "hello"},
            ),
            Case(
                description="too many arguments",
                val=None,
                args=["foo", "bar"],
                kwargs={},
                expect=FilterArgumentError,
            ),
            Case(
                description="missing argument",
                val=None,
                args=[],
                kwargs={},
                expect="",
            ),
            Case(
                description="false returns default",
                val=False,
                args=["bar"],
                kwargs={},
                expect="bar",
            ),
            Case(
                description="empty returns default",
                val=EMPTY,
                args=["bar"],
                kwargs={},
                expect="bar",
            ),
            Case(
                description="allow false",
                val=False,
                args=["bar"],
                kwargs={"allow_false": True},
                expect=False,
            ),
            Case(
                description="undefined",
                val=self.env.undefined("test"),
                args=["bar"],
                kwargs={},
                expect="bar",
            ),
            Case(
                description="a false drop",
                val=MockDrop(False),
                args=["bar"],
                kwargs={},
                expect="bar",
            ),
            Case(
                description="a true drop",
                val=MockDrop(True),
                args=["bar"],
                kwargs={},
                expect=MockDrop(True),
            ),
            Case(
                description="simple drop is always true",
                val=NoLiquidDrop(False),
                args=["bar"],
                kwargs={},
                expect=NoLiquidDrop(False),
            ),
            Case(
                description="zero is not false",
                val=0,
                args=["bar"],
                kwargs={},
                expect=0,
            ),
            Case(
                description="0.0 is not false",
                val=0.0,
                args=["bar"],
                kwargs={},
                expect=0.0,
            ),
            Case(
                description="Decimal('0') is not false",
                val=decimal.Decimal("0"),
                args=["bar"],
                kwargs={},
                expect=decimal.Decimal("0"),
            ),
            Case(
                description="one is not false or true",
                val=1,
                args=["bar"],
                kwargs={},
                expect=1,
            ),
            Case(
                description="falsy drop",
                val=FalsyDrop(0),
                args=["bar"],
                kwargs={},
                expect="bar",
            ),
        ]

        self._test(default, test_cases)

    def test_date(self):
        """Test `date` filter function."""
        test_cases = [
            Case(
                description="datetime object",
                val=datetime.datetime(2002, 1, 1, 11, 45, 13),
                args=[r"%a, %b %d, %y"],
                kwargs={},
                expect="Tue, Jan 01, 02",
            ),
            Case(
                description="well formed string",
                val="March 14, 2016",
                args=[r"%b %d, %y"],
                kwargs={},
                expect="Mar 14, 16",
            ),
            Case(
                description="too many arguments",
                val=datetime.datetime(2001, 2, 1, 11, 45, 13),
                args=[r"%a, %b %d, %y", "foo"],
                kwargs={},
                expect=FilterArgumentError,
            ),
            Case(
                description="not a date or time",
                val=[],
                args=[r"%a, %b %d, %y"],
                kwargs={},
                expect=FilterArgumentError,
            ),
            Case(
                description="missing argument",
                val=None,
                args=[],
                kwargs={},
                expect=FilterArgumentError,
            ),
            Case(
                description="undefined left value",
                val=self.env.undefined("test"),
                args=[r"%b %d, %y"],
                kwargs={},
                expect="",
            ),
            Case(
                description="undefined argument",
                val="March 14, 2016",
                args=[self.env.undefined("test")],
                kwargs={},
                expect="March 14, 2016",
            ),
            Case(
                description="special 'now' value",
                val="now",
                args=["%Y"],
                kwargs={},
                expect=datetime.datetime.now().strftime("%Y"),
            ),
            Case(
                description="special 'today' value",
                val="today",
                args=["%Y"],
                kwargs={},
                expect=datetime.datetime.now().strftime("%Y"),
            ),
            # Case(
            #     description="negative timestamp integer",
            #     val=-1152098955,
            #     args=["%Y"],
            #     kwargs={},
            #     expect="1933",
            # ),
        ]

        self._test(date, test_cases)

    @unittest.skipUnless(platform.system() == "Windows", "windows specific test")
    def test_parse_timestamp_on_windows(self):
        """Test that we can handle parsing timestamps on windows."""
        result = date(-1152098955, "%Y", environment=self.env)
        self.assertEqual(result, "-1152098955")



# pylint: disable=missing-class-docstring missing-module-docstring
import os
import unittest

from mock import patch

from liquid import Template
from liquid import Mode

from liquid.exceptions import LiquidValueError

from liquid.limits import MAX_STR_INT
from liquid.limits import _init_int_max_str_digits


class StrToIntTestCase(unittest.TestCase):
    def test_parse_long_integer_literal(self):
        """Test that we honour the maximum str to int length on integer literals."""
        Template("".join(["{{", "1" * MAX_STR_INT, " }}"]))

        with self.assertRaises(LiquidValueError):
            Template("".join(["{{", "1" * (MAX_STR_INT + 1), " }}"]))

    def test_math_filter_long_string_literal(self):
        """Test that we honour the maximum str to int length on string literals."""
        template = Template("".join(["{{ '", "1" * MAX_STR_INT, "' | abs }}"]))
        self.assertEqual(template.render(), "1" * MAX_STR_INT)

        template = Template(
            "".join(["{{ '", "1" * (MAX_STR_INT + 1), "' | abs }}"]),
            tolerance=Mode.STRICT,
        )

        with self.assertRaises(LiquidValueError):
            self.assertEqual(template.render(), "1" * (MAX_STR_INT + 1))

    @patch.dict(os.environ, {"LIQUIDINTMAXSTRDIGITS": "hello"})
    def test_liquid_max_not_digit(self):
        """Test that we handle liquid environment variables that not digits."""
        with self.assertRaises(TypeError):
            _init_int_max_str_digits()

    @patch.dict(os.environ, {"LIQUIDINTMAXSTRDIGITS": "5"})
    def test_max_digits_too_small(self):
        """Test that we handle environment variables that are too small."""
        with self.assertRaises(ValueError):
            _init_int_max_str_digits()

    @patch.dict(os.environ, {"LIQUIDINTMAXSTRDIGITS": "1000"})
    def test_set_liquid_max(self):
        """Test that we can set LIQUIDINTMAXSTRDIGITS."""
        self.assertEqual(_init_int_max_str_digits(), 1000)

    @patch.dict(os.environ, {"LIQUIDINTMAXSTRDIGITS": "0"})
    def test_zero_is_ok(self):
        """Test that zero is OK"""
        self.assertEqual(_init_int_max_str_digits(), 0)



"""Test array filter functions."""
# pylint: disable=too-many-public-methods,too-many-lines
import unittest

from functools import partial
from inspect import isclass

from typing import Any
from typing import Dict
from typing import Iterable
from typing import List
from typing import NamedTuple


from liquid.environment import Environment
from liquid.expression import NIL


from liquid.exceptions import Error
from liquid.exceptions import FilterError
from liquid.exceptions import FilterArgumentError
from liquid.exceptions import FilterValueError

from liquid.builtin.filters.array import join
from liquid.builtin.filters.array import first
from liquid.builtin.filters.array import last
from liquid.builtin.filters.array import concat
from liquid.builtin.filters.array import map_
from liquid.builtin.filters.array import reverse
from liquid.builtin.filters.array import sort
from liquid.builtin.filters.array import sort_natural
from liquid.builtin.filters.array import where
from liquid.builtin.filters.array import uniq
from liquid.builtin.filters.array import compact

from liquid.extra.filters.array import index


class Case(NamedTuple):
    """Table-driven test helper."""

    description: str
    val: Any
    args: List[Any]
    kwargs: Dict[Any, Any]
    expect: Any


class ArrayFilterTestCase(unittest.TestCase):
    """Test array filter functions."""

    def setUp(self) -> None:
        self.env = Environment()

    def _test(self, func, test_cases: Iterable[Case]):
        if getattr(func, "with_environment", False):
            func = partial(func, environment=self.env)

        for case in test_cases:
            with self.subTest(msg=case.description):
                if isclass(case.expect) and issubclass(case.expect, Error):
                    with self.assertRaises(case.expect):
                        func(case.val, *case.args, **case.kwargs)
                else:
                    self.assertEqual(
                        func(case.val, *case.args, **case.kwargs), case.expect
                    )

    def test_join(self):
        """Test `join` filter function."""

        test_cases = [
            Case(
                description="lists of strings",
                val=["a", "b"],
                args=[
                    "#",
                ],
                kwargs={},
                expect="a#b",
            ),
            Case(
                description="join a string",
                val="a, b",
                args=[
                    "#",
                ],
                kwargs={},
                expect="a, b",
            ),
            Case(
                description="lists of integers",
                val=[1, 2],
                args=[
                    "#",
                ],
                kwargs={},
                expect="1#2",
            ),
            Case(
                description="missing argument defaults to space",
                val=["a", "b"],
                args=[],
                kwargs={},
                expect="a b",
            ),
            Case(
                description="too many arguments",
                val=["a", "b"],
                args=[", ", ""],
                kwargs={},
                expect=FilterArgumentError,
            ),
            Case(
                description="arguments not a string",
                val=["a", "b"],
                args=[5],
                kwargs={},
                expect="a5b",
            ),
            Case(
                description="value not an array",
                val=12,
                args=[", "],
                kwargs={},
                expect="12",
            ),
            Case(
                description="value array contains non string",
                val=["a", "b", 5],
                args=["#"],
                kwargs={},
                expect="a#b#5",
            ),
            Case(
                description="join an undefined variable with a string",
                val=self.env.undefined("test"),
                args=[", "],
                kwargs={},
                expect="",
            ),
            Case(
                description="join an array variable with undefined",
                val=["a", "b"],
                args=[self.env.undefined("test")],
                kwargs={},
                expect="ab",
            ),
        ]

        self._test(join, test_cases)

    def test_first(self):
        """Test `first` filter function."""

        test_cases = [
            Case(
                description="lists of strings",
                val=["a", "b"],
                args=[],
                kwargs={},
                expect="a",
            ),
            Case(
                description="lists of things",
                val=["a", "b", 1, [], {}],
                args=[],
                kwargs={},
                expect="a",
            ),
            Case(
                description="empty list",
                val=[],
                args=[],
                kwargs={},
                expect=None,
            ),
            Case(
                description="unexpected argument",
                val=["a", "b"],
                args=[", "],
                kwargs={},
                expect=FilterArgumentError,
            ),
            Case(
                description="value not an array",
                val=12,
                args=[],
                kwargs={},
                expect=None,
            ),
            Case(
                description="first of undefined",
                val=self.env.undefined("test"),
                args=[],
                kwargs={},
                expect=None,
            ),
        ]

        self._test(first, test_cases)

    def test_last(self):
        """Test `last` filter function."""

        test_cases = [
            Case(
                description="lists of strings",
                val=["a", "b"],
                args=[],
                kwargs={},
                expect="b",
            ),
            Case(
                description="lists of things",
                val=["a", "b", 1, [], {}],
                args=[],
                kwargs={},
                expect={},
            ),
            Case(
                description="empty list",
                val=[],
                args=[],
                kwargs={},
                expect=None,
            ),
            Case(
                description="unexpected argument",
                val=["a", "b"],
                args=[", "],
                kwargs={},
                expect=FilterArgumentError,
            ),
            Case(
                description="value not an array",
                val=12,
                args=[],
                kwargs={},
                expect=None,
            ),
            Case(
                description="last of undefined",
                val=self.env.undefined("test"),
                args=[],
                kwargs={},
                expect=None,
            ),
        ]

        self._test(last, test_cases)

    def test_concat(self):
        """Test `concat` filter function."""

        test_cases = [
            Case(
                description="lists of strings",
                val=["a", "b"],
                args=[["c", "d"]],
                kwargs={},
                expect=["a", "b", "c", "d"],
            ),
            Case(
                description="missing argument",
                val=["a", "b"],
                args=[],
                kwargs={},
                expect=FilterArgumentError,
            ),
            Case(
                description="too many arguments",
                val=["a", "b"],
                args=[["c", "d"], ""],
                kwargs={},
                expect=FilterArgumentError,
            ),
            Case(
                description="arguments not a list",
                val=["a", "b"],
                args=[5],
                kwargs={},
                expect=FilterArgumentError,
            ),
            Case(
                description="not an array",
                val="a, b",
                args=[["c", "d"]],
                kwargs={},
                expect=["a, b", "c", "d"],
            ),
            Case(
                description="array contains non string",
                val=["a", "b", 5],
                args=[["c", "d"]],
                kwargs={},
                expect=["a", "b", 5, "c", "d"],
            ),
            Case(
                description="undefined left value",
                val=self.env.undefined("test"),
                args=[["c", "d"]],
                kwargs={},
                expect=["c", "d"],
            ),
            Case(
                description="undefined argument",
                val=["a", "b"],
                args=[self.env.undefined("test")],
                kwargs={},
                expect=FilterArgumentError,
            ),
        ]

        self._test(concat, test_cases)

    def test_map(self):
        """Test `map` filter function."""

        test_cases = [
            Case(
                description="lists of objects",
                val=[{"title": "foo"}, {"title": "bar"}, {"title": "baz"}],
                args=["title"],
                kwargs={},
                expect=["foo", "bar", "baz"],
            ),
            Case(
                description="missing argument",
                val=[{"title": "foo"}, {"title": "bar"}, {"title": "baz"}],
                args=[],
                kwargs={},
                expect=FilterArgumentError,
            ),
            Case(
                description="too many arguments",
                val=[{"title": "foo"}, {"title": "bar"}, {"title": "baz"}],
                args=["title", ""],
                kwargs={},
                expect=FilterArgumentError,
            ),
            Case(
                description="missing property",
                val=[{"title": "foo"}, {"title": "bar"}, {"heading": "baz"}],
                args=["title"],
                kwargs={},
                expect=["foo", "bar", NIL],
            ),
            Case(
                description="value not an array",
                val=123,
                args=["title"],
                kwargs={},
                expect=FilterError,
            ),
            Case(
                description="array contains non object",
                val=[{"title": "foo"}, {"title": "bar"}, 5, []],
                args=["title"],
                kwargs={},
                expect=FilterError,
            ),
            Case(
                description="undefined left value",
                val=self.env.undefined("test"),
                args=["title"],
                kwargs={},
                expect=[],
            ),
            Case(
                description="undefined argument",
                val=[{"title": "foo"}, {"title": "bar"}, {"title": "baz"}],
                args=[self.env.undefined("test")],
                kwargs={},
                expect=[NIL, NIL, NIL],
            ),
        ]

        self._test(map_, test_cases)

    def test_reverse(self):
        """Test `reverse` filter function."""

        test_cases = [
            Case(
                description="lists of strings",
                val=["b", "a", "B", "A"],
                args=[],
                kwargs={},
                expect=["A", "B", "a", "b"],
            ),
            Case(
                description="lists of things",
                val=["a", "b", 1, [], {}],
                args=[],
                kwargs={},
                expect=[{}, [], 1, "b", "a"],
            ),
            Case(
                description="empty list",
                val=[],
                args=[],
                kwargs={},
                expect=[],
            ),
            Case(
                description="unexpected argument",
                val=["a", "b"],
                args=[", "],
                kwargs={},
                expect=FilterArgumentError,
            ),
            Case(
                description="value not an array",
                val=123,
                args=[],
                kwargs={},
                expect=FilterValueError,
            ),
            Case(
                description="undefined left value",
                val=self.env.undefined("test"),
                args=[],
                kwargs={},
                expect=[],
            ),
        ]

        self._test(reverse, test_cases)

    def test_sort(self):
        """Test `sort` filter function."""

        test_cases = [
            Case(
                description="lists of strings",
                val=["b", "a", "C", "B", "A"],
                args=[],
                kwargs={},
                expect=["A", "B", "C", "a", "b"],
            ),
            Case(
                description="lists of objects with key",
                val=[{"title": "foo"}, {"title": "bar"}, {"title": "Baz"}],
                args=["title"],
                kwargs={},
                expect=[{"title": "Baz"}, {"title": "bar"}, {"title": "foo"}],
            ),
            Case(
                description="lists of objects with missing key",
                val=[{"title": "foo"}, {"title": "bar"}, {"heading": "Baz"}],
                args=["title"],
                kwargs={},
                expect=[{"title": "bar"}, {"title": "foo"}, {"heading": "Baz"}],
            ),
            Case(
                description="empty list",
                val=[],
                args=[],
                kwargs={},
                expect=[],
            ),
            Case(
                description="too many arguments",
                val=[{"title": "foo"}, {"title": "bar"}, {"title": "baz"}],
                args=["title", "heading"],
                kwargs={},
                expect=FilterArgumentError,
            ),
            Case(
                description="value not an array",
                val=123,
                args=[],
                kwargs={},
                expect=FilterValueError,
            ),
            Case(
                description="undefined left value",
                val=self.env.undefined("test"),
                args=[],
                kwargs={},
                expect=[],
            ),
            Case(
                description="undefined argument",
                val=[{"z": "z", "title": "foo"}, {"title": "bar"}, {"title": "Baz"}],
                args=[self.env.undefined("test")],
                kwargs={},
                expect=FilterError,
            ),
            Case(
                description="sort by key targeting an array of strings",
                val=["Z", "b", "a", "C", "A", "B"],
                args=["title"],
                kwargs={},
                expect=["Z", "b", "a", "C", "A", "B"],
            ),
        ]

        self._test(sort, test_cases)

    def test_sort_natural(self):
        """Test `sort_natural` filter function."""

        test_cases = [
            Case(
                description="lists of strings",
                val=["b", "a", "C", "B", "A"],
                args=[],
                kwargs={},
                expect=["a", "A", "b", "B", "C"],
            ),
            Case(
                description="lists of strings with a None",
                val=["b", "a", None, "C", "B", "A"],
                args=[],
                kwargs={},
                expect=["a", "A", "b", "B", "C", None],
            ),
            Case(
                description="lists of objects with key",
                val=[{"title": "foo"}, {"title": "bar"}, {"title": "Baz"}],
                args=["title"],
                kwargs={},
                expect=[{"title": "bar"}, {"title": "Baz"}, {"title": "foo"}],
            ),
            Case(
                description="lists of objects with missing key",
                val=[{"title": "foo"}, {"title": "bar"}, {"heading": "Baz"}],
                args=["title"],
                kwargs={},
                expect=[{"title": "bar"}, {"title": "foo"}, {"heading": "Baz"}],
            ),
            Case(
                description="empty list",
                val=[],
                args=[],
                kwargs={},
                expect=[],
            ),
            Case(
                description="too many arguments",
                val=[{"title": "foo"}, {"title": "bar"}, {"title": "Baz"}],
                args=["title", "heading"],
                kwargs={},
                expect=FilterArgumentError,
            ),
            Case(
                description="value not an array",
                val=1234,
                args=[],
                kwargs={},
                expect=FilterValueError,
            ),
            Case(
                description="undefined left value",
                val=self.env.undefined("test"),
                args=[],
                kwargs={},
                expect=[],
            ),
            Case(
                description="undefined argument",
                val=[{"title": "foo"}, {"title": "bar"}, {"title": "Baz"}],
                args=[self.env.undefined("test")],
                kwargs={},
                expect=[{"title": "bar"}, {"title": "Baz"}, {"title": "foo"}],
            ),
        ]

        self._test(sort_natural, test_cases)

    def test_where(self):
        """Test `where` filter function."""

        test_cases = [
            Case(
                description="lists of object",
                val=[{"title": "foo"}, {"title": "bar"}, {"title": None}],
                args=["title"],
                kwargs={},
                expect=[{"title": "foo"}, {"title": "bar"}],
            ),
            Case(
                description="lists of object with equality test",
                val=[{"title": "foo"}, {"title": "bar"}, {"title": None}],
                args=["title", "bar"],
                kwargs={},
                expect=[{"title": "bar"}],
            ),
            Case(
                description="value not an array",
                val=1234,
                args=["title"],
                kwargs={},
                expect=FilterArgumentError,
            ),
            Case(
                description="missing argument",
                val=[{"title": "foo"}, {"title": "bar"}, {"title": None}],
                args=[],
                kwargs={},
                expect=FilterArgumentError,
            ),
            Case(
                description="too many arguments",
                val=[{"title": "foo"}, {"title": "bar"}, {"title": None}],
                args=["title", "bar", "foo"],
                kwargs={},
                expect=FilterArgumentError,
            ),
            Case(
                description="undefined left value",
                val=self.env.undefined("test"),
                args=["title", "bar"],
                kwargs={},
                expect=[],
            ),
            Case(
                description="undefined first argument",
                val=[{"title": "foo"}, {"title": "bar"}, {"title": None}],
                args=[self.env.undefined("test"), "bar"],
                kwargs={},
                expect=[],
            ),
            Case(
                description="undefined second argument",
                val=[{"title": "foo"}, {"title": "bar"}, {"title": None}],
                args=["title", self.env.undefined("test")],
                kwargs={},
                expect=[{"title": "foo"}, {"title": "bar"}],
            ),
        ]

        self._test(where, test_cases)

    def test_uniq(self):
        """Test `uniq` filter function."""

        test_cases = [
            Case(
                description="lists of strings",
                val=["a", "b", "b", "a"],
                args=[],
                kwargs={},
                expect=["a", "b"],
            ),
            Case(
                description="lists of things",
                val=["a", "b", 1, 1],
                args=[],
                kwargs={},
                expect=["a", "b", 1],
            ),
            Case(
                description="empty list",
                val=[],
                args=[],
                kwargs={},
                expect=[],
            ),
            Case(
                description="unhashable items",
                val=["a", "b", [], {}, {}],
                args=[],
                kwargs={},
                expect=["a", "b", {}],
            ),
            Case(
                description="unexpected argument",
                val=["a", "b"],
                args=["foo", "bar"],
                kwargs={},
                expect=FilterArgumentError,
            ),
            Case(
                description="value not an array",
                val="a, b",
                args=[],
                kwargs={},
                expect=["a, b"],
            ),
            Case(
                description="undefined left value",
                val=self.env.undefined("test"),
                args=[],
                kwargs={},
                expect=[],
            ),
        ]

        self._test(uniq, test_cases)

    def test_compact(self):
        """Test `compact` filter function."""

        test_cases = [
            Case(
                description="lists with nil",
                val=["b", "a", None, "A"],
                args=[],
                kwargs={},
                expect=["b", "a", "A"],
            ),
            Case(
                description="empty list",
                val=[],
                args=[],
                kwargs={},
                expect=[],
            ),
            Case(
                description="unexpected argument",
                val=["a", "b"],
                args=[", "],
                kwargs={},
                expect=FilterArgumentError,
            ),
            Case(
                description="value not an array",
                val=1,
                args=[],
                kwargs={},
                expect=[1],
            ),
            Case(
                description="undefined left value",
                val=self.env.undefined("test"),
                args=[],
                kwargs={},
                expect=[],
            ),
        ]

        self._test(compact, test_cases)

    def test_index(self):
        """Test `index` filter function."""
        test_cases = [
            Case(
                description="array of strings",
                val=["a", "b", "c"],
                args=["b"],
                kwargs={},
                expect=1,
            ),
            Case(
                description="item does not exist",
                val=["a", "b", "c"],
                args=["z"],
                kwargs={},
                expect=None,
            ),
        ]

        self._test(index, test_cases)



# flake8: noqa
# pylint: disable=useless-import-alias,missing-module-docstring

from . import assign_tag
from . import capture_tag
from . import case_tag
from . import comment_tag
from . import cycle_tag
from . import decrement_tag
from . import echo_tag
from . import for_tag
from . import if_tag
from . import ifchanged_tag
from . import illegal
from . import include_tag
from . import increment_tag
from . import liquid_tag
from . import not_liquid
from . import output_statement
from . import raw_tag
from . import render_tag
from . import tablerow_tag
from . import unless_tag
from . import whitespace_control
from . import inline_comment_tag

from . import concat_filter
from . import first_filter
from . import join_filter
from . import last_filter
from . import map_filter
from . import reverse_filter
from . import sort_filter
from . import sort_natural_filter
from . import where_filter
from . import uniq_filter
from . import compact_filter
from . import abs_filter
from . import at_most_filter
from . import at_least_filter
from . import ceil_filter
from . import floor_filter
from . import divided_by_filter
from . import minus_filter
from . import plus_filter
from . import round_filter
from . import times_filter
from . import modulo_filter
from . import size_filter
from . import default_filter
from . import date_filter
from . import capitalize_filter
from . import append_filter
from . import downcase_filter
from . import escape_filter
from . import escape_once_filter
from . import lstrip_filter
from . import newline_to_br_filter
from . import prepend_filter
from . import remove_filter
from . import remove_first_filter
from . import remove_last_filter
from . import replace_filter
from . import replace_first_filter
from . import replace_last_filter
from . import slice_filter
from . import split_filter
from . import upcase_filter
from . import strip_filter
from . import rstrip_filter
from . import strip_html_filter
from . import strip_newlines_filter
from . import truncate_filter
from . import truncatewords_filter
from . import url_encode_filter
from . import url_decode_filter
from . import base64_encode_filter
from . import base64_decode_filter
from . import base64_url_safe_encode_filter
from . import base64_url_safe_decode_filter

from . import range_objects
from . import special
from . import identifiers

test_cases = [
    assign_tag,
    capture_tag,
    case_tag,
    comment_tag,
    cycle_tag,
    decrement_tag,
    echo_tag,
    for_tag,
    if_tag,
    ifchanged_tag,
    illegal,
    include_tag,
    increment_tag,
    liquid_tag,
    not_liquid,
    output_statement,
    raw_tag,
    render_tag,
    tablerow_tag,
    unless_tag,
    whitespace_control,
    concat_filter,
    first_filter,
    join_filter,
    last_filter,
    map_filter,
    reverse_filter,
    sort_filter,
    sort_natural_filter,
    where_filter,
    uniq_filter,
    compact_filter,
    abs_filter,
    at_most_filter,
    at_least_filter,
    ceil_filter,
    floor_filter,
    divided_by_filter,
    minus_filter,
    plus_filter,
    round_filter,
    times_filter,
    modulo_filter,
    size_filter,
    default_filter,
    date_filter,
    capitalize_filter,
    append_filter,
    downcase_filter,
    escape_filter,
    escape_once_filter,
    lstrip_filter,
    newline_to_br_filter,
    prepend_filter,
    remove_filter,
    remove_first_filter,
    remove_last_filter,
    replace_filter,
    replace_first_filter,
    replace_last_filter,
    slice_filter,
    split_filter,
    upcase_filter,
    strip_filter,
    rstrip_filter,
    strip_html_filter,
    strip_newlines_filter,
    truncate_filter,
    truncatewords_filter,
    url_encode_filter,
    url_decode_filter,
    base64_encode_filter,
    base64_decode_filter,
    base64_url_safe_encode_filter,
    base64_url_safe_decode_filter,
    range_objects,
    special,
    inline_comment_tag,
    identifiers,
]



# pylint: disable=missing-class-docstring missing-module-docstring
import unittest
from decimal import Decimal

from typing import Any
from typing import Mapping
from typing import NamedTuple

from liquid import Context
from liquid import Environment

from liquid.expressions import parse_boolean_expression
from liquid.expressions import parse_boolean_expression_with_parens


class Case(NamedTuple):
    """Table-driven test case helper."""

    description: str
    context: Mapping[str, Any]
    expression: str
    expect: Any


class EvalBooleanExpressionTestCase(unittest.TestCase):
    """Test cases for evaluating standard boolean expressions."""

    test_cases = [
        Case(
            description="true literal",
            context={},
            expression="true",
            expect=True,
        ),
        Case(
            description="false literal",
            context={},
            expression="false",
            expect=False,
        ),
        Case(
            description="string literal",
            context={},
            expression="'some'",
            expect=True,
        ),
        Case(
            description="empty string",
            context={},
            expression="''",
            expect=True,
        ),
        Case(
            description="truthy identifier",
            context={"collection": {"title": "foo"}},
            expression="collection.title",
            expect=True,
        ),
        Case(
            description="falsely identifier",
            context={"collection": {"title": "foo"}},
            expression="collection.tags",
            expect=False,
        ),
        Case(
            description="truthy comparison",
            context={"user": {"age": 21}},
            expression="user.age >= 21",
            expect=True,
        ),
        Case(
            description="not equal comparison",
            context={"user": {"age": 21}},
            expression="user.age != 21",
            expect=False,
        ),
        Case(
            description="truthy comparison and logic operator",
            context={
                "user": {"age": 20},
                "collection": {
                    "tags": [
                        "safe",
                    ]
                },
            },
            expression="user.age >= 21 or collection.tags contains 'safe'",
            expect=True,
        ),
        Case(
            description="boolean with logic operators",
            context={},
            expression="true and false and false or true",
            expect=False,
        ),
        Case(
            description="empty array",
            context={"a": {"array": []}},
            expression="a.array == empty",
            expect=True,
        ),
        Case(
            description="empty object",
            context={"a": {"obj": {}}},
            expression="a.obj == empty",
            expect=True,
        ),
        Case(
            description="not empty array",
            context={"a": {"array": [1, 2]}},
            expression="a.array == empty",
            expect=False,
        ),
        Case(
            description="not empty object",
            context={"a": {"obj": {"foo": "bar"}}},
            expression="a.obj == empty",
            expect=False,
        ),
        Case(
            description="invalid comparison to empty",
            context={"a": {"foo": 1}},
            expression="a.foo == empty",
            expect=False,
        ),
        Case(
            description="empty equals empty",
            context={},
            expression="empty == empty",
            expect=True,
        ),
        Case(
            description="empty not equals true",
            context={},
            expression="empty != true",
            expect=True,
        ),
        Case(
            description="nil equals nil",
            context={},
            expression="nil == nil",
            expect=True,
        ),
        Case(
            description="nil equals null",
            context={},
            expression="nil == null",
            expect=True,
        ),
        Case(
            description="null equals null",
            context={},
            expression="null == null",
            expect=True,
        ),
        Case(
            description="string contains string",
            context={},
            expression="'hello' contains 'ell'",
            expect=True,
        ),
        Case(
            description="string contains int",
            context={},
            expression="'hel1lo' contains 1",
            expect=True,
        ),
        Case(
            description="string not equal int",
            context={},
            expression="'hello' != 1",
            expect=True,
        ),
        Case(
            description="array contains",
            context={"foo": [1, 2, 4]},
            expression="foo contains 2",
            expect=True,
        ),
        Case(
            description="array does not contain",
            context={"foo": [1, 2, 4]},
            expression="foo contains 3",
            expect=False,
        ),
        Case(
            description="int equals",
            context={},
            expression="1 == 1",
            expect=True,
        ),
        Case(
            description="int less than",
            context={},
            expression="1 < 2",
            expect=True,
        ),
        Case(
            description="int less than or equal",
            context={},
            expression="1 <= 1",
            expect=True,
        ),
        Case(
            description="int greater than",
            context={},
            expression="1 > 0",
            expect=True,
        ),
        Case(
            description="int greater than or equal",
            context={},
            expression="1 >= 1",
            expect=True,
        ),
        Case(
            description="true equals true",
            context={},
            expression="true == true",
            expect=True,
        ),
        Case(
            description="true equal false",
            context={},
            expression="true == false",
            expect=False,
        ),
        Case(
            description="true not equal false",
            context={},
            expression="true != false",
            expect=True,
        ),
        Case(
            description="string equals int",
            context={},
            expression="'2' == 2",
            expect=False,
        ),
        Case(
            description="empty string is truthy",
            context={},
            expression="''",
            expect=True,
        ),
        Case(
            description="empty string and string is truthy",
            context={},
            expression="'' and 'foo'",
            expect=True,
        ),
        Case(
            description="float equals int",
            context={},
            expression="1 == 1.0",
            expect=True,
        ),
        Case(
            description="empty string equals blank",
            context={},
            expression="'' == blank",
            expect=True,
        ),
        Case(
            description="empty string equals empty",
            context={},
            expression="'' == empty",
            expect=True,
        ),
        Case(
            description="empty array equals empty",
            context={"x": []},
            expression="x == empty",
            expect=True,
        ),
        Case(
            description="empty array equals blank",
            context={"x": []},
            expression="x == blank",
            expect=True,
        ),
        Case(
            description="whitespace string equals blank",
            context={"x": "   "},
            expression="x == blank",
            expect=True,
        ),
        Case(
            description="whitespace string does not equal empty",
            context={"x": "   "},
            expression="x == empty",
            expect=False,
        ),
        Case(
            description="blank does not equal empty",
            context={},
            expression="blank == empty",
            expect=False,
        ),
        Case(
            description="empty does not equal blank",
            context={},
            expression="empty == blank",
            expect=False,
        ),
        Case(
            description="blank does not equal empty",
            context={},
            expression="blank != empty",
            expect=True,
        ),
        Case(
            description="undefined equals nil",
            context={},
            expression="nosuchthing == nil",
            expect=True,
        ),
        Case(
            description="undefined equals null",
            context={},
            expression="nosuchthing == null",
            expect=True,
        ),
        Case(
            description="nil equals undefined",
            context={},
            expression="nil == nosuchthing",
            expect=True,
        ),
        Case(
            description="nil equals None",
            context={"nothing": None},
            expression="nil == nothing",
            expect=True,
        ),
        Case(
            description="nil equals nil",
            context={},
            expression="nil == nil",
            expect=True,
        ),
        Case(
            description="zero",
            context={},
            expression="0",
            expect=True,
        ),
        Case(
            description="0.0",
            context={},
            expression="0.0",
            expect=True,
        ),
        Case(
            description="one",
            context={},
            expression="1",
            expect=True,
        ),
        Case(
            description="zero equals false",
            context={},
            expression="0 == false",
            expect=False,
        ),
        Case(
            description="zero equals true",
            context={},
            expression="0 == true",
            expect=False,
        ),
        Case(
            description="0.0 equals false",
            context={},
            expression="0.0 == false",
            expect=False,
        ),
        Case(
            description="0.0 equals true",
            context={},
            expression="0.0 == true",
            expect=False,
        ),
        Case(
            description="one equals true",
            context={},
            expression="1 == true",
            expect=False,
        ),
        Case(
            description="true equals one",
            context={},
            expression="true == 1",
            expect=False,
        ),
        Case(
            description="one is less than true",
            context={},
            expression="1 < true",
            expect=False,
        ),
        Case(
            description="zero is less than true",
            context={},
            expression="0 < true",
            expect=False,
        ),
        Case(
            description="0.0 is less than true",
            context={},
            expression="0.0 < true",
            expect=False,
        ),
        Case(
            description="one is not equal true",
            context={},
            expression="1 != true",
            expect=True,
        ),
        Case(
            description="zero is not equal false",
            context={},
            expression="0 != false",
            expect=True,
        ),
        Case(
            description="0.0 is not equal false",
            context={},
            expression="0.0 != false",
            expect=True,
        ),
        Case(
            description="false is not equal zero",
            context={},
            expression="false != 0",
            expect=True,
        ),
        Case(
            description="false is not equal 0.0",
            context={},
            expression="false != 0.0",
            expect=True,
        ),
        Case(
            description="false is less than string",
            context={},
            expression="false < 'false'",
            expect=False,
        ),
        Case(
            description="decimal zero",
            context={"n": Decimal("0")},
            expression="n",
            expect=True,
        ),
        Case(
            description="decimal non-zero",
            context={"n": Decimal("1")},
            expression="n",
            expect=True,
        ),
        Case(
            description="decimal zero equals false",
            context={"n": Decimal("0")},
            expression="n == false",
            expect=False,
        ),
        Case(
            description="decimal zero equals true",
            context={"n": Decimal("0")},
            expression="n == true",
            expect=False,
        ),
    ]

    def test_eval_boolean_expression(self) -> None:
        """Test that we can evaluate standard boolean expressions."""
        env = Environment()

        for case in self.test_cases:
            with self.subTest(msg=case.description):
                context = Context(env, globals=case.context)
                expr = parse_boolean_expression(case.expression)
                self.assertEqual(expr.evaluate(context), case.expect)

    def test_eval_boolean_expression_with_parens(self) -> None:
        """Test that we can evaluation of non-standard boolean expressions
        is backwards compatible."""
        env = Environment()

        for case in self.test_cases:
            with self.subTest(msg=case.description):
                context = Context(env, globals=case.context)
                expr = parse_boolean_expression_with_parens(case.expression)
                self.assertEqual(expr.evaluate(context), case.expect)


class EvalBooleanNotExpressionTestCase(unittest.TestCase):
    """Test cases for evaluating non-standard boolean expressions."""

    test_cases = [
        Case(
            description="not true literal",
            context={},
            expression="not true",
            expect=False,
        ),
        Case(
            description="not false literal",
            context={},
            expression="not false",
            expect=True,
        ),
        Case(
            description="not nil literal",
            context={},
            expression="not nil",
            expect=True,
        ),
        Case(
            description="not empty",
            context={},
            expression="not empty",
            expect=False,
        ),
        Case(
            description="not string literal",
            context={},
            expression="not 'some'",
            expect=False,
        ),
        Case(
            description="not empty string",
            context={},
            expression="not ''",
            expect=False,
        ),
        Case(
            description="boolean with logic not operators",
            context={},
            expression="true and not false",
            expect=True,
        ),
        Case(
            description="grouped boolean with logic operators",
            context={},
            expression="(true and false and false) or true",
            expect=True,
        ),
        Case(
            description="nested grouped boolean with logic operators",
            context={},
            expression="((true or false) or (false)) and true",
            expect=True,
        ),
        Case(
            description="grouped boolean with not",
            context={},
            expression="(true and false and false) or not true",
            expect=False,
        ),
        Case(
            description="range literal equals range literal",
            context={},
            expression="(1..3) == (1..3)",
            expect=True,
        ),
        Case(
            description="negate a group",
            context={},
            expression="not (false or true)",
            expect=False,
        ),
    ]

    def test_eval_with_parens(self) -> None:
        """Test that we can evaluate boolean expressions that support logical
        `not` and grouping terms with parentheses."""

        env = Environment()

        for case in self.test_cases:
            with self.subTest(msg=case.description):
                context = Context(env, globals=case.context)
                expr = parse_boolean_expression_with_parens(case.expression)
                self.assertEqual(expr.evaluate(context), case.expect)



"""Test Liquid boolean expression tokenization."""

import unittest

from typing import Any
from typing import NamedTuple

from liquid.exceptions import LiquidSyntaxError

from liquid.expressions.boolean import tokenize
from liquid.expressions.boolean import tokenize_with_parens

from liquid.token import TOKEN_AND
from liquid.token import TOKEN_BLANK
from liquid.token import TOKEN_CONTAINS
from liquid.token import TOKEN_DOT
from liquid.token import TOKEN_EMPTY
from liquid.token import TOKEN_EQ
from liquid.token import TOKEN_FALSE
from liquid.token import TOKEN_GE
from liquid.token import TOKEN_GT
from liquid.token import TOKEN_IDENTIFIER
from liquid.token import TOKEN_INTEGER
from liquid.token import TOKEN_LE
from liquid.token import TOKEN_LG
from liquid.token import TOKEN_LPAREN
from liquid.token import TOKEN_NE
from liquid.token import TOKEN_NIL
from liquid.token import TOKEN_NOT
from liquid.token import TOKEN_NULL
from liquid.token import TOKEN_OR
from liquid.token import TOKEN_RANGE
from liquid.token import TOKEN_RANGE_LITERAL
from liquid.token import TOKEN_RPAREN
from liquid.token import TOKEN_STRING
from liquid.token import TOKEN_TRUE
from liquid.token import Token


class Case(NamedTuple):
    """Table-driven test helper."""

    description: str
    source: str
    expect: Any


class LexBooleanExpressionTestCase(unittest.TestCase):
    """Test cases for the standard boolean expression."""

    test_cases = [
        Case(
            "literal boolean",
            "false == true",
            [
                Token(1, TOKEN_FALSE, "false"),
                Token(1, TOKEN_EQ, "=="),
                Token(1, TOKEN_TRUE, "true"),
            ],
        ),
        Case(
            "not nil identifier",
            "user != nil",
            [
                Token(1, TOKEN_IDENTIFIER, "user"),
                Token(1, TOKEN_NE, "!="),
                Token(1, TOKEN_NIL, "nil"),
            ],
        ),
        Case(
            "not null identifier",
            "user != null",
            [
                Token(1, TOKEN_IDENTIFIER, "user"),
                Token(1, TOKEN_NE, "!="),
                Token(1, TOKEN_NULL, "null"),
            ],
        ),
        Case(
            "alternate not nil",
            "user <> nil",
            [
                Token(1, TOKEN_IDENTIFIER, "user"),
                Token(1, TOKEN_LG, "<>"),
                Token(1, TOKEN_NIL, "nil"),
            ],
        ),
        Case(
            "identifier equals string literal",
            "user.name == 'brian'",
            [
                Token(1, TOKEN_IDENTIFIER, "user"),
                Token(1, TOKEN_DOT, "."),
                Token(1, TOKEN_IDENTIFIER, "name"),
                Token(1, TOKEN_EQ, "=="),
                Token(1, TOKEN_STRING, "brian"),
            ],
        ),
        Case(
            "equality with or",
            "user.name == 'bill' or user.name == 'bob'",
            [
                Token(1, TOKEN_IDENTIFIER, "user"),
                Token(1, TOKEN_DOT, "."),
                Token(1, TOKEN_IDENTIFIER, "name"),
                Token(1, TOKEN_EQ, "=="),
                Token(1, TOKEN_STRING, "bill"),
                Token(1, TOKEN_OR, "or"),
                Token(1, TOKEN_IDENTIFIER, "user"),
                Token(1, TOKEN_DOT, "."),
                Token(1, TOKEN_IDENTIFIER, "name"),
                Token(1, TOKEN_EQ, "=="),
                Token(1, TOKEN_STRING, "bob"),
            ],
        ),
        Case(
            "equality with and",
            "user.name == 'bob' and user.age > 45",
            [
                Token(1, TOKEN_IDENTIFIER, "user"),
                Token(1, TOKEN_DOT, "."),
                Token(1, TOKEN_IDENTIFIER, "name"),
                Token(1, TOKEN_EQ, "=="),
                Token(1, TOKEN_STRING, "bob"),
                Token(1, TOKEN_AND, "and"),
                Token(1, TOKEN_IDENTIFIER, "user"),
                Token(1, TOKEN_DOT, "."),
                Token(1, TOKEN_IDENTIFIER, "age"),
                Token(1, TOKEN_GT, ">"),
                Token(1, TOKEN_INTEGER, "45"),
            ],
        ),
        Case(
            "greater than or equal to integer literal",
            "user.age >= 21",
            [
                Token(1, TOKEN_IDENTIFIER, "user"),
                Token(1, TOKEN_DOT, "."),
                Token(1, TOKEN_IDENTIFIER, "age"),
                Token(1, TOKEN_GE, ">="),
                Token(1, TOKEN_INTEGER, "21"),
            ],
        ),
        Case(
            "less than or equal to integer literal",
            "user.age <= 21",
            [
                Token(1, TOKEN_IDENTIFIER, "user"),
                Token(1, TOKEN_DOT, "."),
                Token(1, TOKEN_IDENTIFIER, "age"),
                Token(1, TOKEN_LE, "<="),
                Token(1, TOKEN_INTEGER, "21"),
            ],
        ),
        Case(
            "identifier contains string",
            "product.tags contains 'sale'",
            [
                Token(1, TOKEN_IDENTIFIER, "product"),
                Token(1, TOKEN_DOT, "."),
                Token(1, TOKEN_IDENTIFIER, "tags"),
                Token(1, TOKEN_CONTAINS, "contains"),
                Token(1, TOKEN_STRING, "sale"),
            ],
        ),
        Case(
            "identifier equals blank",
            "product.title == blank",
            [
                Token(1, TOKEN_IDENTIFIER, "product"),
                Token(1, TOKEN_DOT, "."),
                Token(1, TOKEN_IDENTIFIER, "title"),
                Token(1, TOKEN_EQ, "=="),
                Token(1, TOKEN_BLANK, "blank"),
            ],
        ),
        Case(
            "identifier equals empty",
            "product.title == empty",
            [
                Token(1, TOKEN_IDENTIFIER, "product"),
                Token(1, TOKEN_DOT, "."),
                Token(1, TOKEN_IDENTIFIER, "title"),
                Token(1, TOKEN_EQ, "=="),
                Token(1, TOKEN_EMPTY, "empty"),
            ],
        ),
    ]

    def test_default_lexer(self) -> None:
        """Test that we can tokenize standard boolean expressions."""
        for case in self.test_cases:
            with self.subTest(msg=case.description):
                tokens = list(tokenize(case.source))
                self.assertEqual(len(tokens), len(case.expect))

                for got, want in zip(tokens, case.expect):
                    self.assertEqual(got, want)

    def test_lexer_with_parens(self) -> None:
        """Test that the non-standard boolean expression lexer is backwards
        compatible with standard boolean expressions."""
        for case in self.test_cases:
            with self.subTest(msg=case.description):
                tokens = list(tokenize_with_parens(case.source))
                self.assertEqual(len(tokens), len(case.expect))

                for got, want in zip(tokens, case.expect):
                    self.assertEqual(got, want)

    def test_not(self) -> None:
        """Test that the `not` keyword is not recognized in default boolean
        expressions."""
        source = "not true"
        tokens = list(tokenize(source))
        self.assertEqual(len(tokens), 2)
        self.assertEqual(
            tokens,
            [
                Token(1, TOKEN_IDENTIFIER, "not"),  # ident, not TOKEN_NOT
                Token(1, TOKEN_TRUE, "true"),
            ],
        )


class LexBooleanNotExpressionTestCase(unittest.TestCase):
    """Test cases for the non-standard boolean expression that supports
    logical `not` and grouping with parentheses."""

    test_cases = [
        Case(
            "literal boolean not true",
            "false == not true",
            [
                Token(1, TOKEN_FALSE, "false"),
                Token(1, TOKEN_EQ, "=="),
                Token(1, TOKEN_NOT, "not"),
                Token(1, TOKEN_TRUE, "true"),
            ],
        ),
        Case(
            "literal boolean not false",
            "false == not false",
            [
                Token(1, TOKEN_FALSE, "false"),
                Token(1, TOKEN_EQ, "=="),
                Token(1, TOKEN_NOT, "not"),
                Token(1, TOKEN_FALSE, "false"),
            ],
        ),
        Case(
            "parens",
            "(false and false)",
            [
                Token(1, TOKEN_LPAREN, "("),
                Token(1, TOKEN_FALSE, "false"),
                Token(1, TOKEN_AND, "and"),
                Token(1, TOKEN_FALSE, "false"),
                Token(1, TOKEN_RPAREN, ")"),
            ],
        ),
        Case(
            "range literals",
            "(1..3) == (1..3)",
            [
                Token(1, TOKEN_RANGE_LITERAL, "rangeliteral"),
                Token(1, TOKEN_LPAREN, "("),
                Token(1, TOKEN_INTEGER, "1"),
                Token(1, TOKEN_RANGE, ".."),
                Token(1, TOKEN_INTEGER, "3"),
                Token(1, TOKEN_RPAREN, ")"),
                Token(1, TOKEN_EQ, "=="),
                Token(1, TOKEN_RANGE_LITERAL, "rangeliteral"),
                Token(1, TOKEN_LPAREN, "("),
                Token(1, TOKEN_INTEGER, "1"),
                Token(1, TOKEN_RANGE, ".."),
                Token(1, TOKEN_INTEGER, "3"),
                Token(1, TOKEN_RPAREN, ")"),
            ],
        ),
        Case(
            "nested grouped boolean with logic operators",
            "((true or false) or (false)) and true",
            [
                Token(1, TOKEN_LPAREN, "("),
                Token(1, TOKEN_LPAREN, "("),
                Token(1, TOKEN_TRUE, "true"),
                Token(1, TOKEN_OR, "or"),
                Token(1, TOKEN_FALSE, "false"),
                Token(1, TOKEN_RPAREN, ")"),
                Token(1, TOKEN_OR, "or"),
                Token(1, TOKEN_LPAREN, "("),
                Token(1, TOKEN_FALSE, "false"),
                Token(1, TOKEN_RPAREN, ")"),
                Token(1, TOKEN_RPAREN, ")"),
                Token(1, TOKEN_AND, "and"),
                Token(1, TOKEN_TRUE, "true"),
            ],
        ),
    ]

    def test_lexer_with_parens(self) -> None:
        """Test that we can tokenize boolean expressions that support logical
        `not` and grouping with parentheses."""
        for case in self.test_cases:
            with self.subTest(msg=case.description):
                tokens = list(tokenize_with_parens(case.source))
                self.assertEqual(len(tokens), len(case.expect))

                for got, want in zip(tokens, case.expect):
                    self.assertEqual(got, want)


class LexBooleanExpressionErrorsTestCase(unittest.TestCase):
    """Test cases for tokens that are illegal in a boolean expression."""

    test_cases = [
        Case(
            description="illegal characters",
            source="} == %",
            expect=LiquidSyntaxError,
        ),
        Case(
            description="unknown operator",
            source="x >< y",
            expect=LiquidSyntaxError,
        ),
    ]

    def test_lex_illegal_expression(self) -> None:
        """Test that we raise a syntax error upon illegal characters."""
        for case in self.test_cases:
            with self.subTest(msg=case.description):
                with self.assertRaises(case.expect):
                    list(tokenize(case.source))

    def test_lex_illegal_expression_with_parens(self) -> None:
        """Test that we raise a syntax error upon illegal characters when using the
        non-standard lexer."""
        for case in self.test_cases:
            with self.subTest(msg=case.description):
                with self.assertRaises(case.expect):
                    list(tokenize_with_parens(case.source))



"""Test cases for shorthand comment syntax."""
# pylint: disable=missing-class-docstring

from dataclasses import dataclass
from unittest import TestCase

from typing import List
from typing import Type
from typing import Union

from liquid import Environment
from liquid import Template
from liquid import Mode


@dataclass
class Case:
    description: str
    template: str
    expect: Union[str, Type[Exception]]


class ShorthandCommentTestCase(TestCase):
    def _test(
        self,
        test_cases: List[Case],
        template_comments=False,
        comment_start_string="{#",
        comment_end_string="#}",
    ) -> None:
        """Table driven test runner."""
        for case in test_cases:
            with self.subTest(msg=case.description, env=True):
                env = Environment(
                    tolerance=Mode.STRICT,
                    template_comments=template_comments,
                    comment_start_string=comment_start_string,
                    comment_end_string=comment_end_string,
                )

                template = env.from_string(case.template)
                result = template.render()
                self.assertEqual(result, case.expect)

            # Test using the Template API too.
            with self.subTest(msg=case.description, Template=True):
                template = Template(
                    case.template,
                    tolerance=Mode.STRICT,
                    template_comments=template_comments,
                    comment_start_string=comment_start_string,
                    comment_end_string=comment_end_string,
                )
                result = template.render()
                self.assertEqual(result, case.expect)

    def test_no_shorthand_comment_syntax(self):
        """Test that a default environment does not support shorthand comment syntax."""
        test_cases = [
            Case(
                description="shorthand comments are disbled",
                template=r"Hello, {# this is a comment #} World!",
                expect="Hello, {# this is a comment #} World!",
            ),
        ]

        self._test(test_cases)

    def test_shorthand_comment_syntax(self):
        """Test that shorthand comments are ignored."""
        test_cases = [
            Case(
                description="shorthand comments are ignored",
                template=r"Hello, {# this is a comment #} World!",
                expect="Hello,  World!",
            ),
            Case(
                description="shorthand comments must have a trailing hash.",
                template=r"Hello, {# this is not a comment } World!",
                expect="Hello, {# this is not a comment } World!",
            ),
            Case(
                description="shorthand comments can contain tags and statements",
                template=r"Hello, {# this is {% if %} a {{ comment }} #} World!",
                expect="Hello,  World!",
            ),
            Case(
                description="liquid expression line comments are ignored",
                template="\n".join(
                    (
                        r"{% liquid ",
                        r"if true",
                        r"echo 'foo'",
                        r"endif",
                        r"# this is a comment",
                        r"%}",
                    )
                ),
                expect="foo",
            ),
        ]

        self._test(test_cases, template_comments=True)

    def test_shorthand_comment_whitespace_control(self):
        """Test that shorthand comments can use whitespace control."""
        test_cases = [
            Case(
                description="shorthand comments can remove trailing whitespace",
                template=r"Hello, {# this is a comment -#} World!",
                expect="Hello, World!",
            ),
            Case(
                description="shorthand comments can remove leading whitespace",
                template=r"Hello, {#- this is a comment #} World!",
                expect="Hello, World!",
            ),
        ]

        self._test(test_cases, template_comments=True)

    def test_different_start_and_end_sequence(self):
        """Test that we can change the comment start and end sequence."""
        test_cases = [
            Case(
                description="shorthand comments are ignored",
                template="Hello, {// this is a comment //} World!",
                expect="Hello,  World!",
            ),
            Case(
                description="shorthand comments must have a valid end sequence.",
                template=r"Hello, {// this is not a comment } World!",
                expect="Hello, {// this is not a comment } World!",
            ),
            Case(
                description="shorthand comments can contain tags and statements",
                template=r"Hello, {// this is {% if %} a {{ comment }} //} World!",
                expect="Hello,  World!",
            ),
            Case(
                description="liquid expression line comments are ignored",
                template="\n".join(
                    (
                        r"{% liquid ",
                        r"if true",
                        r"echo 'foo'",
                        r"endif",
                        r"// this is a comment",
                        r"%}",
                    )
                ),
                expect="foo",
            ),
        ]

        self._test(
            test_cases,
            template_comments=True,
            comment_start_string="{//",
            comment_end_string="//}",
        )



"""Golden tests cases for testing liquid's built-in `truncate` filter."""

from liquid.golden.case import Case

cases = [
    Case(
        description="default end",
        template=r'{{ "Ground control to Major Tom." | truncate: 20 }}',
        expect="Ground control to...",
    ),
    Case(
        description="custom end",
        template=r'{{ "Ground control to Major Tom." | truncate: 25, ", and so on" }}',
        expect="Ground control, and so on",
    ),
    Case(
        description="no end",
        template=r'{{ "Ground control to Major Tom." | truncate: 20, "" }}',
        expect="Ground control to Ma",
    ),
    Case(
        description="string is shorter than length",
        template=r'{{ "Ground control" | truncate: 20 }}',
        expect="Ground control",
    ),
    Case(
        description="not a string",
        template=r"{{ 5 | truncate: 10 }}",
        expect="5",
    ),
    Case(
        description="too many arguments",
        template=r'{{ "hello" | truncate: 5, "foo", "bar" }}',
        expect="",
        error=True,
    ),
    Case(
        description="undefined left value",
        template=r"{{ nosuchthing | truncate: 5 }}",
        expect="",
    ),
    Case(
        description="undefined first argument",
        template=r'{{ "Ground control to Major Tom." | truncate: nosuchthing }}',
        expect="",
        error=True,
    ),
    Case(
        description="undefined second argument",
        template=r'{{ "Ground control to Major Tom." | truncate: 20, nosuchthing }}',
        expect="Ground control to Ma",
    ),
    Case(
        description="default length is 50",
        template=(
            r'{{ "Ground control to Major Tom. Ground control to Major Tom." '
            r"| truncate }}"
        ),
        expect="Ground control to Major Tom. Ground control to ...",
    ),
]



"""Golden tests cases for testing liquid's built-in `ceil` filter."""

from liquid.golden.case import Case

cases = [
    Case(
        description="positive integer",
        template=r"{{ 5 | floor }}",
        expect="5",
    ),
    Case(
        description="negative integer",
        template=r"{{ -5 | floor }}",
        expect="-5",
    ),
    Case(
        description="positive float",
        template=r"{{ 5.4 | floor }}",
        expect="5",
    ),
    Case(
        description="negative float",
        template=r"{{ -5.4 | floor }}",
        expect="-6",
    ),
    Case(
        description="zero",
        template=r"{{ 0 | floor }}",
        expect="0",
    ),
    Case(
        description="positive string float",
        template=r'{{ "5.1" | floor }}',
        expect="5",
    ),
    Case(
        description="negative string float",
        template=r'{{ "-5.1" | floor }}',
        expect="-6",
    ),
    Case(
        description="unexpected argument",
        template=r"{{ -3.1 | floor: 1 }}",
        expect="",
        error=True,
    ),
    Case(
        description="string not a number",
        template=r'{{ "hello" | floor }}',
        expect="0",
    ),
    Case(
        description="not a string, int or float",
        template=r"{{ a | floor }}",
        expect="0",
        globals={"a": {}},
    ),
    Case(
        description="undefined left value",
        template=r"{{ nosuchthing | floor }}",
        expect="0",
    ),
]



"""Test cases for statically counting a template's variables."""
import asyncio

from typing import List, Optional, TextIO
from unittest import TestCase

from liquid import Environment
from liquid import DictLoader
from liquid import Template

from liquid.ast import ChildNode, Node
from liquid.context import Context
from liquid.exceptions import TemplateTraversalError
from liquid.expression import Expression
from liquid.extra import add_inline_expression_tags
from liquid.extra import WithTag
from liquid.mode import Mode
from liquid.stream import TokenStream
from liquid.tag import Tag
from liquid.template import BoundTemplate, Refs
from liquid.token import Token


class MockExpression(Expression):
    """Mock expression"""

    def evaluate(self, context: Context) -> object:
        return "mock expression"

    async def evaluate_async(self, context: Context) -> object:
        return "mock expression"


class MockNode(Node):
    """Mock AST node."""

    def __init__(self, token: Token) -> None:
        self.tok = token

    def render_to_output(self, context: Context, buffer: TextIO) -> Optional[bool]:
        buffer.write("mock node")

    async def render_to_output_async(
        self, context: Context, buffer: TextIO
    ) -> Optional[bool]:
        buffer.write("mock node")


class MockTag(Tag):
    """Mock tag."""

    block = False
    name = "mock"

    def parse(self, stream: TokenStream) -> Node:
        return MockNode(stream.current)


class MockChildNode(MockNode):
    """Mock AST node."""

    def children(self) -> List[ChildNode]:
        return [ChildNode(self.tok.linenum, expression=MockExpression())]


class MockChildTag(MockTag):
    """Mock tag"""

    def parse(self, stream: TokenStream) -> Node:
        return MockChildNode(stream.current)


# pylint: disable=too-many-public-methods
class CountTemplateVariablesTestCase(TestCase):
    """Test that we can count a template's variable references."""

    # pylint: disable=too-many-arguments
    def _test(
        self,
        template: BoundTemplate,
        template_refs: Refs,
        template_locals: Refs,
        template_globals: Refs,
        failed_visits: Optional[Refs] = None,
        unloadable: Optional[Refs] = None,
        raise_for_failures: bool = True,
    ) -> None:
        failed_visits = failed_visits if failed_visits is not None else {}
        unloadable = unloadable if unloadable is not None else {}

        refs = template.analyze(raise_for_failures=raise_for_failures)
        self.assertEqual(refs.local_variables, template_locals)
        self.assertEqual(refs.global_variables, template_globals)
        self.assertEqual(refs.variables, template_refs)
        self.assertEqual(refs.failed_visits, failed_visits)
        self.assertEqual(refs.unloadable_partials, unloadable)

        async def coro():
            return await template.analyze_async(raise_for_failures=raise_for_failures)

        refs = asyncio.run(coro())
        self.assertEqual(refs.local_variables, template_locals)
        self.assertEqual(refs.global_variables, template_globals)
        self.assertEqual(refs.variables, template_refs)
        self.assertEqual(refs.failed_visits, failed_visits)
        self.assertEqual(refs.unloadable_partials, unloadable)

    def test_analyze_output(self):
        """Test that we can count references in an output statement."""
        template = Template("{{ x | default: y, allow_false: z }}")

        expected_template_globals = {
            "x": [("<string>", 1)],
            "y": [("<string>", 1)],
            "z": [("<string>", 1)],
        }
        expected_template_locals = {}
        expected_refs = {
            "x": [("<string>", 1)],
            "y": [("<string>", 1)],
            "z": [("<string>", 1)],
        }

        self._test(
            template,
            expected_refs,
            expected_template_locals,
            expected_template_globals,
        )

    def test_analyze_identifier_with_bracketed_string_literal(self):
        """Test that we can count references in bracketed identifiers."""
        template = Template("{{ x['y'].title }}")

        expected_template_globals = {"x.y.title": [("<string>", 1)]}
        expected_template_locals = {}
        expected_refs = {"x.y.title": [("<string>", 1)]}

        self._test(
            template,
            expected_refs,
            expected_template_locals,
            expected_template_globals,
        )

    def test_analyze_identifier_with_nested_identifier(self):
        """Test that we can count references in nested identifiers."""
        template = Template("{{ x[y.z].title }}")

        expected_template_globals = {
            "x.[y.z].title": [("<string>", 1)],
            "y.z": [("<string>", 1)],
        }
        expected_template_locals = {}
        expected_refs = {"x.[y.z].title": [("<string>", 1)], "y.z": [("<string>", 1)]}

        self._test(
            template,
            expected_refs,
            expected_template_locals,
            expected_template_globals,
        )

    def test_analyze_assign_tag(self):
        """Test that we can count references in an assign tag."""
        template = Template("{% assign x = y | append: z %}")

        expected_template_globals = {"y": [("<string>", 1)], "z": [("<string>", 1)]}
        expected_template_locals = {"x": [("<string>", 1)]}
        expected_refs = {"y": [("<string>", 1)], "z": [("<string>", 1)]}

        self._test(
            template,
            expected_refs,
            expected_template_locals,
            expected_template_globals,
        )

    def test_analyze_capture_tag(self):
        """Test that we can count references in a capture tag."""
        template = Template("{% capture x %}{% if y %}z{% endif %}{% endcapture %}")

        expected_template_globals = {"y": [("<string>", 1)]}
        expected_template_locals = {"x": [("<string>", 1)]}
        expected_refs = {"y": [("<string>", 1)]}

        self._test(
            template,
            expected_refs,
            expected_template_locals,
            expected_template_globals,
        )

    def test_analyze_case_tag(self):
        """Test that we can count references in a case tag."""
        # NOTE: `case` tags get rewritten to `if`/`elsif` tags, resulting in unintuitive
        # line numbers when analyzing `case` tags.
        template = Template(
            "\n".join(
                [
                    "{% case x %}",
                    "{% when y %}",
                    "  {{ a }}",
                    "{% when z %}",
                    "  {{ b }}",
                    "{% endcase %}",
                ]
            )
        )

        expected_template_globals = {
            "x": [("<string>", 2), ("<string>", 4)],
            "y": [("<string>", 2)],
            "a": [("<string>", 3)],
            "z": [("<string>", 4)],
            "b": [("<string>", 5)],
        }
        expected_template_locals = {}
        expected_refs = {
            "x": [("<string>", 2), ("<string>", 4)],
            "y": [("<string>", 2)],
            "a": [("<string>", 3)],
            "z": [("<string>", 4)],
            "b": [("<string>", 5)],
        }

        self._test(
            template,
            expected_refs,
            expected_template_locals,
            expected_template_globals,
        )

    def test_analyze_cycle_tag(self):
        """Test that we can count references in a cycle tag."""
        template = Template("{% cycle x: a, b %}")

        expected_template_globals = {
            "x": [("<string>", 1)],
            "a": [("<string>", 1)],
            "b": [("<string>", 1)],
        }
        expected_template_locals = {}
        expected_refs = {
            "x": [("<string>", 1)],
            "a": [("<string>", 1)],
            "b": [("<string>", 1)],
        }

        self._test(
            template,
            expected_refs,
            expected_template_locals,
            expected_template_globals,
        )

    def test_analyze_decrement_tag(self):
        """Test that we can count references in a decrement tag."""
        template = Template("{% decrement x %}")

        expected_template_globals = {}
        expected_template_locals = {"x": [("<string>", 1)]}
        expected_refs = {}

        self._test(
            template,
            expected_refs,
            expected_template_locals,
            expected_template_globals,
        )

    def test_analyze_echo_tag(self):
        """Test that we can count references in an echo tag."""
        template = Template(r"{% echo x | default: y, allow_false: z %}")

        expected_template_globals = {
            "x": [("<string>", 1)],
            "y": [("<string>", 1)],
            "z": [("<string>", 1)],
        }
        expected_template_locals = {}
        expected_refs = {
            "x": [("<string>", 1)],
            "y": [("<string>", 1)],
            "z": [("<string>", 1)],
        }

        self._test(
            template,
            expected_refs,
            expected_template_locals,
            expected_template_globals,
        )

    def test_analyze_for_tag(self):
        """Test that we can count references in a for tag."""
        template = Template(
            "\n".join(
                [
                    "{% for x in (1..y) %}",
                    "  {{ x }}",
                    "{% else %}",
                    "  {{ z }}",
                    "{% endfor %}",
                ]
            )
        )

        expected_template_globals = {"y": [("<string>", 1)], "z": [("<string>", 4)]}
        expected_template_locals = {}
        expected_refs = {
            "x": [("<string>", 2)],
            "y": [("<string>", 1)],
            "z": [("<string>", 4)],
        }

        self._test(
            template,
            expected_refs,
            expected_template_locals,
            expected_template_globals,
        )

    def test_analyze_if_tag(self):
        """Test that we can count references in an if tag."""
        template = Template(
            "\n".join(
                [
                    "{% if x %}",
                    "  {{ a }}",
                    "{% elsif y %}",
                    "  {{ b }}",
                    "{% endif %}",
                ]
            )
        )

        expected_template_globals = {
            "x": [("<string>", 1)],
            "a": [("<string>", 2)],
            "y": [("<string>", 3)],
            "b": [("<string>", 4)],
        }
        expected_template_locals = {}
        expected_refs = {
            "x": [("<string>", 1)],
            "a": [("<string>", 2)],
            "y": [("<string>", 3)],
            "b": [("<string>", 4)],
        }

        self._test(
            template,
            expected_refs,
            expected_template_locals,
            expected_template_globals,
        )

    def test_analyze_ifchanged_tag(self):
        """Test that we can count references in an ifchanged tag."""
        template = Template(r"{% ifchanged %}{{ x }}{% endifchanged %}")

        expected_template_globals = {"x": [("<string>", 1)]}
        expected_template_locals = {}
        expected_refs = {"x": [("<string>", 1)]}

        self._test(
            template,
            expected_refs,
            expected_template_locals,
            expected_template_globals,
        )

    def test_analyze_increment_tag(self):
        """Test that we can count references in an increment tag."""
        template = Template("{% increment x %}")

        expected_template_globals = {}
        expected_template_locals = {"x": [("<string>", 1)]}
        expected_refs = {}

        self._test(
            template,
            expected_refs,
            expected_template_locals,
            expected_template_globals,
        )

    def test_analyze_liquid_tag(self):
        """Test that we can count references in a liquid tag."""
        template = Template(
            "\n".join(
                [
                    r"{% liquid",
                    r"if product.title",
                    r"   echo foo | upcase",
                    r"else",
                    r"   echo 'product-1' | upcase ",
                    r"endif",
                    r"",
                    r"for i in (0..5)",
                    r"   echo i",
                    r"endfor %}",
                ]
            )
        )

        expected_template_globals = {
            "product.title": [("<string>", 2)],
            "foo": [("<string>", 3)],
        }
        expected_template_locals = {}
        expected_refs = {
            "product.title": [("<string>", 2)],
            "foo": [("<string>", 3)],
            "i": [("<string>", 9)],
        }

        self._test(
            template,
            expected_refs,
            expected_template_locals,
            expected_template_globals,
        )

    def test_analyze_tablerow_tag(self):
        """Test that we can count references in a tablerow tag."""
        template = Template(
            r"{% tablerow x in y.z %}{{ x | append: a }}{% endtablerow %}",
        )

        expected_template_globals = {"y.z": [("<string>", 1)], "a": [("<string>", 1)]}
        expected_template_locals = {}
        expected_refs = {
            "y.z": [("<string>", 1)],
            "x": [("<string>", 1)],
            "a": [("<string>", 1)],
        }

        self._test(
            template,
            expected_refs,
            expected_template_locals,
            expected_template_globals,
        )

    def test_analyze_unless_tag(self):
        """Test that we can count references in an unless tag."""
        template = Template(
            "\n".join(
                [
                    "{% unless x %}",
                    "  {{ a }}",
                    "{% elsif y %}",
                    "  {{ b }}",
                    "{% endunless %}",
                ]
            )
        )

        expected_template_globals = {
            "x": [("<string>", 1)],
            "a": [("<string>", 2)],
            "y": [("<string>", 3)],
            "b": [("<string>", 4)],
        }
        expected_template_locals = {}
        expected_refs = {
            "x": [("<string>", 1)],
            "a": [("<string>", 2)],
            "y": [("<string>", 3)],
            "b": [("<string>", 4)],
        }

        self._test(
            template,
            expected_refs,
            expected_template_locals,
            expected_template_globals,
        )

    def test_analyze_include_tag(self):
        """Test that we can count references in an include tag."""
        loader = DictLoader({"some_name": "{{ y }}"})
        env = Environment(loader=loader)
        template = env.from_string("{% include 'some_name' %}")

        expected_template_globals = {"y": [("some_name", 1)]}
        expected_template_locals = {}
        expected_refs = {"y": [("some_name", 1)]}

        self._test(
            template,
            expected_refs,
            expected_template_locals,
            expected_template_globals,
        )

    def test_analyze_include_tag_with_assign(self):
        """Test that we can count references in an include tag and update template
        locals ."""
        loader = DictLoader({"some_name": "{{ y }}{% assign z = 4 %}"})
        env = Environment(loader=loader)
        template = env.from_string("{% include 'some_name' %}{{ z }}")

        expected_template_globals = {"y": [("some_name", 1)]}
        expected_template_locals = {"z": [("some_name", 1)]}
        expected_refs = {"y": [("some_name", 1)], "z": [("<string>", 1)]}

        self._test(
            template,
            expected_refs,
            expected_template_locals,
            expected_template_globals,
        )

    def test_analyze_include_tag_once(self):
        """Test that we analyze each partial template once."""
        loader = DictLoader({"some_name": "{{ y }}"})
        env = Environment(loader=loader)
        template = env.from_string("{% include 'some_name' %}{% include 'some_name' %}")

        expected_template_globals = {"y": [("some_name", 1)]}
        expected_template_locals = {}
        expected_refs = {"y": [("some_name", 1)]}

        self._test(
            template,
            expected_refs,
            expected_template_locals,
            expected_template_globals,
        )

    def test_analyze_recursive_include_tag(self):
        """Test that we can analyze recursive include tags."""
        loader = DictLoader({"some_name": "{{ y }}{% include 'some_name' %}"})
        env = Environment(loader=loader)
        template = env.from_string("{% include 'some_name' %}")

        expected_template_globals = {"y": [("some_name", 1)]}
        expected_template_locals = {}
        expected_refs = {"y": [("some_name", 1)]}

        self._test(
            template,
            expected_refs,
            expected_template_locals,
            expected_template_globals,
        )

    def test_analyze_include_tag_with_bound_variable(self):
        """Test that we can count references in an include tag that uses a bound
        variable."""
        # Without an "as" alias, bound variables are given the name of the included
        # template, up to the first dot.
        loader = DictLoader({"some_name": "{{ y | append: x }}{{ some_name }}"})
        env = Environment(loader=loader)
        template = env.from_string("{% include 'some_name' with foo %}")

        expected_template_globals = {
            "foo": [("<string>", 1)],
            "y": [("some_name", 1)],
            "x": [("some_name", 1)],
        }
        expected_template_locals = {}
        expected_refs = {
            "y": [("some_name", 1)],
            "foo": [("<string>", 1)],
            "x": [("some_name", 1)],
            "some_name": [("some_name", 1)],
        }

        self._test(
            template,
            expected_refs,
            expected_template_locals,
            expected_template_globals,
        )

    def test_analyze_include_tag_with_alias(self):
        """Test that we can count references in an include tag that uses an alias."""
        loader = DictLoader({"some_name": "{{ y | append: x }}"})
        env = Environment(loader=loader)
        template = env.from_string("{% include 'some_name' with foo as x %}")

        expected_template_globals = {"foo": [("<string>", 1)], "y": [("some_name", 1)]}
        expected_template_locals = {}
        expected_refs = {
            "y": [("some_name", 1)],
            "foo": [("<string>", 1)],
            "x": [("some_name", 1)],
        }

        self._test(
            template,
            expected_refs,
            expected_template_locals,
            expected_template_globals,
        )

    def test_analyze_include_tag_with_arguments(self):
        """Test that we can count references in an include tag with arguments."""
        loader = DictLoader({"some_name": "{{ y | append: x }}"})
        env = Environment(loader=loader)
        template = env.from_string("{% include 'some_name', x:y, z:'4' %}\n{{ x }}")

        expected_template_globals = {
            "y": [("some_name", 1), ("<string>", 1)],
            "x": [("<string>", 2)],
        }
        expected_template_locals = {}
        expected_refs = {
            "x": [("some_name", 1), ("<string>", 2)],
            "y": [("some_name", 1), ("<string>", 1)],
        }

        self._test(
            template,
            expected_refs,
            expected_template_locals,
            expected_template_globals,
        )

    def test_include_with_variable_name(self):
        """Test that we handle include tags with a variable template name."""
        env = Environment()
        template = env.from_string("{% include somevar %}{{ y }}")

        expected_template_globals = {
            "somevar": [("<string>", 1)],
            "y": [("<string>", 1)],
        }
        expected_template_locals = {}
        expected_refs = {"somevar": [("<string>", 1)], "y": [("<string>", 1)]}
        expected_unloadable_partials = {"somevar": [("<string>", 1)]}

        self._test(
            template,
            expected_refs,
            expected_template_locals,
            expected_template_globals,
            unloadable=expected_unloadable_partials,
            raise_for_failures=False,
        )

        with self.assertRaises(TemplateTraversalError):
            self._test(
                template,
                expected_refs,
                expected_template_locals,
                expected_template_globals,
                unloadable=expected_unloadable_partials,
                raise_for_failures=True,
            )

    def test_include_template_not_found(self):
        """Test that we handle templates that can not be found."""
        env = Environment()
        template = env.from_string("{% include 'nosuchtemplate' %}{{ y }}")

        expected_template_globals = {"y": [("<string>", 1)]}
        expected_template_locals = {}
        expected_refs = {"y": [("<string>", 1)]}
        expected_unloadable_partials = {"nosuchtemplate": [("<string>", 1)]}

        self._test(
            template,
            expected_refs,
            expected_template_locals,
            expected_template_globals,
            unloadable=expected_unloadable_partials,
            raise_for_failures=False,
        )

        with self.assertRaises(TemplateTraversalError):
            self._test(
                template,
                expected_refs,
                expected_template_locals,
                expected_template_globals,
                unloadable=expected_unloadable_partials,
                raise_for_failures=True,
            )

    def test_analyze_render_tag(self):
        """Test that we can count references in a render tag."""
        loader = DictLoader({"some_name": "{{ x }}{% assign y = z %}"})
        env = Environment(loader=loader)
        template = env.from_string("{% assign z = 1 %}{% render 'some_name' %}")

        expected_template_globals = {"x": [("some_name", 1)], "z": [("some_name", 1)]}
        expected_template_locals = {"z": [("<string>", 1)]}
        expected_refs = {"x": [("some_name", 1)], "z": [("some_name", 1)]}

        self._test(
            template,
            expected_refs,
            expected_template_locals,
            expected_template_globals,
        )

    def test_analyze_render_tag_once(self):
        """Test that analyze each partial rendered template once."""
        loader = DictLoader({"some_name": "{{ x }}"})
        env = Environment(loader=loader)
        template = env.from_string("{% render 'some_name' %}{% render 'some_name' %}")

        expected_template_globals = {"x": [("some_name", 1)]}
        expected_template_locals = {}
        expected_refs = {"x": [("some_name", 1)]}

        self._test(
            template,
            expected_refs,
            expected_template_locals,
            expected_template_globals,
        )

    def test_analyze_recursive_render_tag(self):
        """Test that we can analyze recursive render tags."""
        loader = DictLoader({"some_name": "{{ y }}{% render 'some_name' %}"})
        env = Environment(loader=loader)
        template = env.from_string("{% render 'some_name' %}")

        expected_template_globals = {"y": [("some_name", 1)]}
        expected_template_locals = {}
        expected_refs = {"y": [("some_name", 1)]}

        self._test(
            template,
            expected_refs,
            expected_template_locals,
            expected_template_globals,
        )

    def test_analyze_render_tag_with_bound_variable(self):
        """Test that we can count references in a render tag that uses a bound
        variable."""
        loader = DictLoader({"some_name": "{{ y | append: x }}{{ some_name }}"})
        env = Environment(loader=loader)
        template = env.from_string("{% render 'some_name' with foo %}")

        expected_template_globals = {
            "foo": [("<string>", 1)],
            "y": [("some_name", 1)],
            "x": [("some_name", 1)],
        }
        expected_template_locals = {}
        expected_refs = {
            "y": [("some_name", 1)],
            "foo": [("<string>", 1)],
            "x": [("some_name", 1)],
            "some_name": [("some_name", 1)],
        }

        self._test(
            template,
            expected_refs,
            expected_template_locals,
            expected_template_globals,
        )

    def test_analyze_render_tag_with_alias(self):
        """Test that we can count references in a render tag that uses an alias."""
        loader = DictLoader({"some_name": "{{ y | append: x }}"})
        env = Environment(loader=loader)
        template = env.from_string("{% render 'some_name' with foo as x %}")

        expected_template_globals = {"foo": [("<string>", 1)], "y": [("some_name", 1)]}
        expected_template_locals = {}
        expected_refs = {
            "y": [("some_name", 1)],
            "foo": [("<string>", 1)],
            "x": [("some_name", 1)],
        }

        self._test(
            template,
            expected_refs,
            expected_template_locals,
            expected_template_globals,
        )

    def test_analyze_render_tag_with_arguments(self):
        """Test that we can count references in a render tag with arguments."""
        loader = DictLoader({"some_name": "{{ y | append: x }}"})
        env = Environment(loader=loader)
        template = env.from_string("{% render 'some_name', x:y, z:'4' %}\n{{ x }}")

        expected_template_globals = {
            "y": [("some_name", 1), ("<string>", 1)],
            "x": [("<string>", 2)],
        }
        expected_template_locals = {}
        expected_refs = {
            "x": [("some_name", 1), ("<string>", 2)],
            "y": [("some_name", 1), ("<string>", 1)],
        }

        self._test(
            template,
            expected_refs,
            expected_template_locals,
            expected_template_globals,
        )

    def test_analyze_render_tag_scope(self):
        """Test that we local references in a render tag go out of scope."""
        loader = DictLoader({"some_name": "{{ x }}{% assign y = z %}"})
        env = Environment(loader=loader)
        template = env.from_string("{% assign z = 1 %}{% render 'some_name' %}{{ y }}")

        expected_template_globals = {
            "x": [("some_name", 1)],
            "z": [("some_name", 1)],
            "y": [("<string>", 1)],
        }
        expected_template_locals = {"z": [("<string>", 1)]}
        expected_refs = {
            "x": [("some_name", 1)],
            "z": [("some_name", 1)],
            "y": [("<string>", 1)],
        }

        self._test(
            template,
            expected_refs,
            expected_template_locals,
            expected_template_globals,
        )

    def test_render_template_not_found(self):
        """Test that we handle templates that can not be found."""
        env = Environment()
        template = env.from_string("{% render 'nosuchtemplate' %}{{ y }}")

        expected_template_globals = {"y": [("<string>", 1)]}
        expected_template_locals = {}
        expected_refs = {"y": [("<string>", 1)]}
        expected_unloadable_partials = {"nosuchtemplate": [("<string>", 1)]}

        self._test(
            template,
            expected_refs,
            expected_template_locals,
            expected_template_globals,
            unloadable=expected_unloadable_partials,
            raise_for_failures=False,
        )

        with self.assertRaises(TemplateTraversalError):
            self._test(
                template,
                expected_refs,
                expected_template_locals,
                expected_template_globals,
                unloadable=expected_unloadable_partials,
                raise_for_failures=True,
            )

    def test_node_missing_children(self):
        """Test that we handle AST nodes that don't implement a children method."""
        env = Environment(tolerance=Mode.STRICT)
        env.add_tag(MockTag)
        template = env.from_string("{% mock %}\n{% mock %}")

        expected_failed_visits = {"MockNode": [("<string>", 1), ("<string>", 2)]}

        self._test(
            template,
            {},
            {},
            {},
            failed_visits=expected_failed_visits,
            raise_for_failures=False,
        )

        with self.assertRaises(TemplateTraversalError):
            self._test(
                template,
                {},
                {},
                {},
                failed_visits=expected_failed_visits,
                raise_for_failures=True,
            )

    def test_expression_missing_children(self):
        """Test that we handle expressions that don't implement a children method."""
        env = Environment(tolerance=Mode.STRICT)
        env.add_tag(MockChildTag)
        template = env.from_string("{% mock %}\n{% mock %}")

        expected_failed_visits = {"MockExpression": [("<string>", 1), ("<string>", 2)]}

        self._test(
            template,
            {},
            {},
            {},
            failed_visits=expected_failed_visits,
            raise_for_failures=False,
        )

        with self.assertRaises(TemplateTraversalError):
            self._test(
                template,
                {},
                {},
                {},
                failed_visits=expected_failed_visits,
                raise_for_failures=True,
            )

    def test_analyze_conditional_expression(self) -> None:
        """Test that we can statically analyze non-standard conditional expressions."""
        env = Environment()
        add_inline_expression_tags(env)

        template = env.from_string(r"{{ foo | upcase if a.b else bar || append: baz }}")

        expected_template_globals = {
            "foo": [("<string>", 1)],
            "a.b": [("<string>", 1)],
            "bar": [("<string>", 1)],
            "baz": [("<string>", 1)],
        }
        expected_template_locals: Refs = {}
        expected_refs = {
            "foo": [("<string>", 1)],
            "a.b": [("<string>", 1)],
            "bar": [("<string>", 1)],
            "baz": [("<string>", 1)],
        }

        self._test(
            template,
            expected_refs,
            expected_template_locals,
            expected_template_globals,
        )

    def test_analyze_with_tag(self) -> None:
        """Test that we can statically analyze non-standard with tags."""
        env = Environment()
        env.add_tag(WithTag)

        template = env.from_string(
            r"{% with p: collection.products.first %}"
            r"{{ p.title }}"
            r"{% endwith %}"
            r"{{ p.title }}"
            r"{{ collection.products.first.title }}"
        )

        expected_template_globals = {
            "collection.products.first": [("<string>", 1)],
            "p.title": [("<string>", 1)],
            "collection.products.first.title": [("<string>", 1)],
        }
        expected_template_locals: Refs = {}
        expected_refs = {
            "collection.products.first": [("<string>", 1)],
            "collection.products.first.title": [("<string>", 1)],
            "p.title": [("<string>", 1), ("<string>", 1)],
        }

        self._test(
            template,
            expected_refs,
            expected_template_locals,
            expected_template_globals,
        )



"""Parse tree node and tag definition for the built in "include" tag."""
import sys

from typing import List
from typing import Optional
from typing import Dict
from typing import TextIO
from typing import Tuple

from liquid.ast import ChildNode
from liquid.ast import Node

from liquid.builtin.drops import IterableDrop
from liquid.context import Context

from liquid.expression import Expression, Literal
from liquid.expression import Identifier

from liquid.exceptions import LiquidSyntaxError
from liquid.stream import TokenStream
from liquid.tag import Tag

from liquid.parse import expect

from liquid.token import Token
from liquid.token import TOKEN_EXPRESSION
from liquid.token import TOKEN_IDENTIFIER
from liquid.token import TOKEN_WITH
from liquid.token import TOKEN_FOR
from liquid.token import TOKEN_AS
from liquid.token import TOKEN_COMMA
from liquid.token import TOKEN_COLON
from liquid.token import TOKEN_EOF

from liquid.expressions import TokenStream as ExprTokenStream
from liquid.expressions.common import parse_string_or_identifier
from liquid.expressions.common import parse_unchained_identifier
from liquid.expressions.common import parse_identifier
from liquid.expressions.include.lex import tokenize
from liquid.expressions.filtered.parse import parse_obj


TAG_INCLUDE = sys.intern("include")


class IncludeNode(Node):
    """Parse tree node for the built-in "include" tag."""

    __slots__ = ("tok", "name", "var", "alias", "args")
    tag = TAG_INCLUDE

    def __init__(
        self,
        tok: Token,
        name: Expression,
        var: Optional[Identifier] = None,
        alias: Optional[str] = None,
        args: Optional[Dict[str, Expression]] = None,
    ):
        self.tok = tok
        self.name = name
        self.var = var
        self.alias = alias
        self.args = args or {}

    def __str__(self) -> str:
        buf = [f"{self.name}"]

        if self.var:
            buf.append(f" with {self.var}")

        if self.alias:
            buf.append(f" as {self.alias}")

        if self.args:
            buf.append(", ")

        args = (f"{key}={val}" for key, val in self.args.items())
        buf.append(", ".join(args))

        return f"{self.tag}({''.join(buf)})"

    def __repr__(self) -> str:
        return f"IncludeNode(tok={self.tok!r}, name={self.name})"  # pragma: no cover

    def render_to_output(self, context: Context, buffer: TextIO) -> Optional[bool]:
        name = self.name.evaluate(context)
        template = context.get_template_with_context(str(name), tag=self.tag)

        namespace: Dict[str, object] = {}

        # Add any keyword arguments to the new template context.
        for _key, _val in self.args.items():
            namespace[_key] = _val.evaluate(context)

        with context.extend(namespace):
            # Bind a variable to the included template.
            if self.var is not None:
                val = self.var.evaluate(context)
                key = self.alias or template.name.split(".")[0]

                # If the variable is array-like, render the template once for each
                # item in the array.
                #
                # The reference implementation does not seem to distinguish between
                # "for" and "with". Just checks for array-like-ness.
                if isinstance(val, (tuple, list, IterableDrop)):
                    # NOTE: What if an included template with a bound array updates
                    # a keyword argument value? Do we need to update namespace
                    # arguments after each loop?
                    #
                    # The reference implementation seems to evaluate arguments once,
                    # before the loop.
                    context.raise_for_loop_limit(len(val))
                    for itm in val:
                        namespace[key] = itm
                        template.render_with_context(context, buffer, partial=True)
                else:
                    namespace[key] = val
                    template.render_with_context(context, buffer, partial=True)
            else:
                template.render_with_context(context, buffer, partial=True)

        return True

    async def render_to_output_async(
        self, context: Context, buffer: TextIO
    ) -> Optional[bool]:
        """Same as ``render_to_output`` but uses async versions of get_template and
        render_with_context."""
        name = await self.name.evaluate_async(context)
        template = await context.get_template_with_context_async(
            str(name), tag=self.tag
        )
        namespace: Dict[str, object] = {}

        for _key, _val in self.args.items():
            namespace[_key] = await _val.evaluate_async(context)

        with context.extend(namespace):
            if self.var is not None:
                val = await self.var.evaluate_async(context)
                key = self.alias or template.name.split(".")[0]

                if isinstance(val, (tuple, list, IterableDrop)):
                    context.raise_for_loop_limit(len(val))
                    for itm in val:
                        namespace[key] = itm
                        await template.render_with_context_async(
                            context, buffer, partial=True
                        )
                else:
                    namespace[key] = val
                    await template.render_with_context_async(
                        context, buffer, partial=True
                    )
            else:
                await template.render_with_context_async(context, buffer, partial=True)

        return True

    def children(self) -> List[ChildNode]:
        block_scope: List[str] = list(self.args.keys())
        _children = [
            ChildNode(
                linenum=self.tok.linenum,
                node=None,
                expression=self.name,
                block_scope=block_scope,
                load_mode="include",
                load_context={"tag": "include"},
            )
        ]
        if self.var:
            if self.alias:
                block_scope.append(self.alias)
            elif isinstance(self.name, Literal):
                block_scope.append(str(self.name.value).split(".", 1)[0])
            _children.append(
                ChildNode(
                    linenum=self.tok.linenum,
                    expression=self.var,
                )
            )
        for expr in self.args.values():
            _children.append(ChildNode(linenum=self.tok.linenum, expression=expr))
        return _children


BIND_TOKENS = frozenset((TOKEN_WITH, TOKEN_FOR))


class IncludeTag(Tag):
    """The built-in "include" tag."""

    name = TAG_INCLUDE
    block = False
    node_class = IncludeNode

    def parse(self, stream: TokenStream) -> Node:
        """Read an IncludeNode from the given stream of tokens."""
        tok = next(stream)
        expect(stream, TOKEN_EXPRESSION)
        expr_stream = ExprTokenStream(
            tokenize(
                stream.current.value,
                linenum=tok.linenum,
            )
        )

        # Need a string or identifier that resolves to a string. This is the name
        # of the template to be included.
        name = parse_string_or_identifier(expr_stream)
        next(expr_stream)

        identifier: Optional[Identifier] = None
        alias: Optional[str] = None

        # Optionally bind a variable to the included template context
        if expr_stream.current[1] in BIND_TOKENS:
            next(expr_stream)  # Eat 'with' or 'for'
            expr_stream.expect(TOKEN_IDENTIFIER)
            identifier = parse_identifier(expr_stream)
            next(expr_stream)

            # The bound variable will take the name of the template by default,
            # or an alias if an identifier follows the "as" keyword.
            if expr_stream.current[1] == TOKEN_AS:
                next(expr_stream)  # Eat 'as'
                expr_stream.expect(TOKEN_IDENTIFIER)
                alias = str(parse_unchained_identifier(expr_stream))
                next(expr_stream)

        # Zero or more keyword arguments
        args: Dict[str, Expression] = {}

        # The first keyword argument might follow immediately or after a comma.
        if expr_stream.current[1] == TOKEN_IDENTIFIER:
            key, val = _parse_argument(expr_stream)
            args[key] = val

        while expr_stream.current[1] != TOKEN_EOF:
            if expr_stream.current[1] == TOKEN_COMMA:
                next(expr_stream)  # Eat comma
                key, val = _parse_argument(expr_stream)
                args[key] = val
            else:
                typ = expr_stream.current[1]
                raise LiquidSyntaxError(
                    f"expected a comma separated list of arguments, found {typ}",
                    linenum=tok.linenum,
                )

        return self.node_class(tok, name=name, var=identifier, alias=alias, args=args)


def _parse_argument(stream: ExprTokenStream) -> Tuple[str, Expression]:
    key = str(parse_unchained_identifier(stream))
    stream.next_token()
    stream.expect(TOKEN_COLON)
    stream.next_token()  # Eat colon
    val = parse_obj(stream)
    stream.next_token()
    return key, val



"""Golden tests cases for testing liquid's built-in `truncatewords` filter."""

from liquid.golden.case import Case

cases = [
    Case(
        description="default end",
        template=r'{{ "Ground control to Major Tom." | truncatewords: 3 }}',
        expect="Ground control to...",
    ),
    Case(
        description="custom end",
        template=r'{{ "Ground control to Major Tom." | truncatewords: 3, "--" }}',
        expect="Ground control to--",
    ),
    Case(
        description="no end",
        template=r'{{ "Ground control to Major Tom." | truncatewords: 3, "" }}',
        expect="Ground control to",
    ),
    Case(
        description="fewer words than word count",
        template=r'{{ "Ground control" | truncatewords: 3 }}',
        expect="Ground control",
    ),
    Case(
        description="not a string",
        template="{{ 5 | truncatewords: 10 }}",
        expect="5",
    ),
    Case(
        description="too many arguments",
        template='{{ "hello" | truncatewords: 5, "foo", "bar" }}',
        expect="",
        error=True,
    ),
    Case(
        description="reference implementation test 1",
        template=r'{{ "测试测试测试测试" | truncatewords: 5 }}',
        expect="测试测试测试测试",
    ),
    Case(
        description="reference implementation test 2",
        template=r'{{ "one two three" | truncatewords: 2, 1 }}',
        expect="one two1",
    ),
    Case(
        description="reference implementation test 3",
        template='{{ "one  two\tthree\nfour" | truncatewords: 3 }}',
        expect="one two three...",
    ),
    Case(
        description="reference implementation test 4",
        template=r'{{ "one two three four" | truncatewords: 2 }}',
        expect="one two...",
    ),
    Case(
        description="reference implementation test 5",
        template=r'{{ "one two three four" | truncatewords: 0 }}',
        expect="one...",
    ),
    Case(
        description="undefined left value",
        template=r"{{ nosuchthing | truncatewords: 5 }}",
        expect="",
    ),
    Case(
        description="undefined first argument",
        template=r'{{ "one two three four" | truncatewords: nosuchthing }}',
        expect="",
        error=True,
    ),
    Case(
        description="undefined second argument",
        template=r'{{ "one two three four" | truncatewords: 2, nosuchthing }}',
        expect="one two",
    ),
    # Case(
    #     description="very big argument",
    #     template=r'{{ "" | truncatewords: 100000000000000 }}',
    #     expect="",
    #     error=False,
    # ),
    Case(
        description="number of words defaults to 15",
        template=r'{{ "a b c d e f g h i j k l m n o p q" | truncatewords }}',
        expect="a b c d e f g h i j k l m n o...",
    ),
    Case(
        description="all whitespace is clobbered",
        template=r'{{ "    one    two three    four  " | truncatewords: 2 }}',
        expect="one two...",
    ),
]



"""Golden tests cases for testing liquid's built-in `map` filter."""

from liquid.golden.case import Case

cases = [
    Case(
        description="array of objects",
        template=r"{{ a | map: 'title' | join: '#' }}",
        expect="foo#bar#baz",
        globals={"a": [{"title": "foo"}, {"title": "bar"}, {"title": "baz"}]},
    ),
    Case(
        description="missing property",
        template=r"{{ a | map: 'title' | join: '#' }}",
        expect="foo#bar#",
        globals={"a": [{"title": "foo"}, {"title": "bar"}, {"heading": "baz"}]},
    ),
    Case(
        description="left value not an array",
        template=r"{{ a | map: 'title' | join: '#' }}",
        expect="",
        globals={"a": 123},
        error=True,
    ),
    Case(
        description="array containing a non object",
        template=r"{{ a | map: 'title' | join: '#' }}",
        expect="",
        globals={"a": [{"title": "foo"}, {"title": "bar"}, 5, []]},
        error=True,
    ),
    Case(
        description="undefined argument",
        template=r"{{ a | map: nosuchthing | join: '#' }}",
        expect="#",
        globals={"a": [{"title": "foo"}, {"title": "bar"}]},
    ),
]



"""Tag and node definition for the built-in "unless" tag."""
from __future__ import annotations
import sys

from typing import List
from typing import Optional
from typing import TextIO
from typing import TYPE_CHECKING
from typing import Union

from liquid.ast import ChildNode
from liquid.ast import Node
from liquid.ast import BlockNode
from liquid.ast import ConditionalBlockNode
from liquid.ast import IllegalNode

from liquid.context import Context
from liquid.exceptions import LiquidSyntaxError
from liquid.stream import TokenStream
from liquid.tag import Tag

from liquid.parse import eat_block
from liquid.parse import expect
from liquid.parse import get_parser

from liquid.token import Token
from liquid.token import TOKEN_EOF
from liquid.token import TOKEN_EXPRESSION
from liquid.token import TOKEN_TAG

if TYPE_CHECKING:  # pragma: no cover
    from liquid import Environment
    from liquid.expression import Expression

TAG_UNLESS = sys.intern("unless")
TAG_ENDUNLESS = sys.intern("endunless")
TAG_ELSIF = sys.intern("elsif")
TAG_ELSE = sys.intern("else")

ENDUNLESSBLOCK = frozenset((TAG_ENDUNLESS, TAG_ELSIF, TAG_ELSE, TOKEN_EOF))
ENDELSIFBLOCK = frozenset((TAG_ENDUNLESS, TAG_ELSIF, TAG_ELSE))
ENDUNLESSELSEBLOCK = frozenset((TAG_ENDUNLESS,))


class UnlessNode(Node):
    """Parse tree node for the built-in "unless" tag."""

    __slots__ = (
        "tok",
        "condition",
        "consequence",
        "conditional_alternatives",
        "alternative",
        "forced_output",
    )

    def __init__(
        self,
        tok: Token,
        condition: Expression,
        consequence: BlockNode,
        conditional_alternatives: Optional[List[ConditionalBlockNode]] = None,
        alternative: Optional[BlockNode] = None,
    ):
        self.tok = tok
        self.condition = condition
        self.consequence = consequence
        self.conditional_alternatives = conditional_alternatives or []
        self.alternative = alternative

        self.forced_output = any(
            getattr(n, "forced_output", False)
            for n in (
                self.consequence,
                *self.conditional_alternatives,
                self.alternative,
            )
            if n
        )

    def __str__(self) -> str:
        buf = [
            f"if !{self.condition} {{ {self.consequence} }}",
        ]

        for alt in self.conditional_alternatives:
            buf.append(f"elsif {alt}")

        if self.alternative:
            buf.append(f"else {{ {self.alternative} }}")
        return " ".join(buf)

    def render_to_output(self, context: Context, buffer: TextIO) -> Optional[bool]:
        # This intermediate buffer is used to detect and possibly suppress blocks that,
        # when rendered, contain only whitespace.
        buf = context.get_buffer(buffer)

        if not self.condition.evaluate(context):
            rendered = self.consequence.render(context, buf)
        else:
            rendered = False
            for alt in self.conditional_alternatives:
                if alt.render(context, buf):
                    rendered = True
                    break

            if not rendered and self.alternative:
                rendered = self.alternative.render(context, buf)

        val = buf.getvalue()
        if self.forced_output or not val.isspace():
            buffer.write(val)

        return rendered

    async def render_to_output_async(
        self, context: Context, buffer: TextIO
    ) -> Optional[bool]:
        # This intermediate buffer is used to detect and possibly suppress blocks that,
        # when rendered, contain only whitespace.
        buf = context.get_buffer(buffer)

        if not await self.condition.evaluate_async(context):
            rendered = await self.consequence.render_async(context, buf)
        else:
            rendered = False
            for alt in self.conditional_alternatives:
                if await alt.render_async(context, buf):
                    rendered = True
                    break

            if not rendered and self.alternative:
                rendered = await self.alternative.render_async(context, buf)

        val = buf.getvalue()
        if self.forced_output or not val.isspace():
            buffer.write(val)

        return rendered

    def children(self) -> List[ChildNode]:
        _children = [
            ChildNode(
                linenum=self.consequence.tok.linenum,
                node=self.consequence,
                expression=self.condition,
            )
        ]
        _children.extend(
            [
                ChildNode(
                    linenum=alt.tok.linenum,
                    node=alt.block,
                    expression=alt.condition,
                )
                for alt in self.conditional_alternatives
            ]
        )
        if self.alternative:
            _children.append(
                ChildNode(
                    linenum=self.alternative.tok.linenum,
                    node=self.alternative,
                    expression=None,
                )
            )
        return _children


class UnlessTag(Tag):
    """The built-in "unless" tag."""

    name = TAG_UNLESS
    end = TAG_ENDUNLESS

    def __init__(self, env: Environment):
        super().__init__(env)
        self.parser = get_parser(self.env)

    def parse_expression(self, stream: TokenStream) -> Expression:
        """Parse a boolean expression from a stream of tokens."""
        expect(stream, TOKEN_EXPRESSION)
        return self.env.parse_boolean_expression_value(stream.current.value)

    def parse(self, stream: TokenStream) -> Union[UnlessNode, IllegalNode]:

        expect(stream, TOKEN_TAG, value=TAG_UNLESS)
        tok = stream.current
        stream.next_token()

        condition = self.parse_expression(stream)
        stream.next_token()

        consequence = self.parser.parse_block(stream, ENDUNLESSBLOCK)

        conditional_alternatives = []

        while stream.current.istag(TAG_ELSIF):
            stream.next_token()

            try:
                expr = self.parse_expression(stream)
            except LiquidSyntaxError as err:
                self.env.error(err)
                eat_block(stream, ENDELSIFBLOCK)
                return IllegalNode(tok)

            alt_tok = stream.current
            stream.next_token()
            alt_block = self.parser.parse_block(stream, ENDELSIFBLOCK)

            conditional_alternatives.append(
                ConditionalBlockNode(alt_tok, condition=expr, block=alt_block)
            )

        alternative: Optional[BlockNode] = None

        if stream.current.istag(TAG_ELSE):
            stream.next_token()
            alternative = self.parser.parse_block(stream, ENDUNLESSELSEBLOCK)

        expect(stream, TOKEN_TAG, value=TAG_ENDUNLESS)
        return UnlessNode(
            tok,
            condition=condition,
            consequence=consequence,
            conditional_alternatives=conditional_alternatives,
            alternative=alternative,
        )



"""Token stream test cases."""
from unittest import TestCase

from typing import List

from liquid.token import TOKEN_STRING
from liquid.token import TOKEN_PIPE
from liquid.token import TOKEN_IDENTIFIER
from liquid.token import TOKEN_COLON
from liquid.token import TOKEN_COMMA
from liquid.token import TOKEN_INTEGER
from liquid.token import TOKEN_EOF

from liquid.exceptions import LiquidSyntaxError

from liquid.expressions import Token
from liquid.expressions import TokenStream


class ExpresssionTokenStreamTestCase(TestCase):
    """Test the expression token stream."""

    def setUp(self) -> None:
        self.tokens: List[Token] = [
            (1, TOKEN_STRING, "Liquid"),
            (1, TOKEN_PIPE, "|"),
            (1, TOKEN_IDENTIFIER, "slice"),
            (1, TOKEN_COLON, ":"),
            (1, TOKEN_INTEGER, "2"),
            (1, TOKEN_COMMA, ","),
            (1, TOKEN_INTEGER, "5"),
        ]

    def test_iterate_stream(self):
        """Test that we can iterate a token stream."""
        stream = TokenStream(iter(self.tokens))

        for i, token in enumerate(stream):
            self.assertEqual(token, self.tokens[i])

        stream = TokenStream(iter(self.tokens))
        stream_iter = iter(stream)

        self.assertEqual(list(stream_iter), self.tokens)

    def test_step_through_stream(self):
        """Test that we can step through a token stream."""
        stream = TokenStream(iter(self.tokens))

        for token in self.tokens:
            self.assertEqual(token, stream.next_token())

    def test_step_through_stream_with_next(self):
        """Test that we can step through a token stream using next."""
        stream = TokenStream(iter(self.tokens))

        for token in self.tokens:
            self.assertEqual(token, next(stream))

    def test_end_of_stream(self):
        """Test that we get an EOF token when the stream is exhausted."""
        stream = TokenStream(iter(self.tokens))
        _ = list(stream)
        self.assertEqual(stream.current[1], TOKEN_EOF)
        self.assertEqual(next(stream)[1], TOKEN_EOF)

    def test_peek(self):
        """Test that we can peek at the next token."""
        stream = TokenStream(iter(self.tokens))

        for i in range(len(self.tokens) - 1):
            self.assertEqual(stream.peek, self.tokens[i + 1])
            next(stream)

    def test_push(self):
        """Test that we can push tokens back onto the stream."""
        stream = TokenStream(iter(self.tokens))

        self.assertEqual(stream.current, self.tokens[0])
        self.assertEqual(stream.peek, self.tokens[1])

        token = next(stream)
        self.assertEqual(token, self.tokens[0])
        self.assertEqual(stream.current, self.tokens[1])
        self.assertEqual(stream.peek, self.tokens[2])

        stream.push(token)
        self.assertEqual(next(stream), self.tokens[0])

        stream.push(token)
        self.assertEqual(stream.current, self.tokens[0])
        self.assertEqual(stream.peek, self.tokens[1])

    def test_expect(self):
        """Test that we can make assertions about the current token."""
        stream = TokenStream(iter(self.tokens))
        stream.expect(TOKEN_STRING)

        with self.assertRaises(LiquidSyntaxError):
            stream.expect(TOKEN_INTEGER)

    def test_expect_peek(self):
        """Test that we can make assertions about the next token."""
        stream = TokenStream(iter(self.tokens))
        stream.expect_peek(TOKEN_PIPE)

        with self.assertRaises(LiquidSyntaxError):
            stream.expect_peek(TOKEN_INTEGER)
